# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of AtomsProperties that relate to NMR J
couplings"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from soprano.properties import AtomsProperty
from soprano.selection import AtomSelection
from soprano.nmr.utils import (_haeb_sort, _anisotropy, _asymmetry,
                               _span, _skew, _evecs_2_quat,
                               _J_constant)
from soprano.data.nmr import _get_nmr_data, _get_isotope_data


def _has_isc_check(f):
    # Decorator to add a check for the J coupling array
    def decorated_f(s, *args, **kwargs):
        tag = kwargs.get('tag', 'isc')
        if not (s.has(tag)):
            raise RuntimeError('No J coupling data for tag {0}'.format(tag) +
                               ' found in this system')
        return f(s, *args, **kwargs)

    return decorated_f


class ISCDiagonal(AtomsProperty):

    """
    ISCDiagonal

    Produces an array containing eigenvalues and eigenvectors for the
    symmetric part of each J coupling tensor in the system. By default
    saves them as part of the Atoms' arrays as well.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   tag (str): name of the J coupling component to return. Magres files
    |              usually contain isc, isc_spin, isc_fc, isc_orbital_p and
    |              isc_orbital_d. Default is isc.
    |   save_info (bool): if True, save the diagonalised tensors in the
    |                     Atoms object as info. By default True.

    | Returns:
    |   isc_diag (dict): dictionary of eigenvalues and eigenvectors by atom
    |                    index pairs (lower index first)

    """

    default_name = 'isc_diagonal'
    default_params = {
        'tag': 'isc',
        'save_info': True
    }

    @staticmethod
    @_has_isc_check
    def extract(s, tag, save_info):

        isc_dict = {(i, j): np.array(t) for j, r in enumerate(s.get_array(tag))
                    for i, t in enumerate(r) if t is not None}

        isc_diag = {ij: dict(zip(['evals', 'evecs'],
                                 np.linalg.eigh((t+t.T)/2.0)))
                    for ij, t in isc_dict.items()}
        isc_pairs = sorted(isc_diag.keys())
        isc_evals = np.array([isc_diag[ij]['evals'] for ij in isc_pairs])
        isc_evecs = np.array([isc_diag[ij]['evecs'] for ij in isc_pairs])

        if save_info:
            s.info[ISCDiagonal.default_name + '_' +
                   tag + '_pairs'] = isc_pairs
            s.info[ISCDiagonal.default_name + '_' +
                   tag + '_evals'] = isc_evals
            s.info[ISCDiagonal.default_name + '_' +
                   tag + '_evals_hsort'] = _haeb_sort(isc_evals)
            s.info[ISCDiagonal.default_name + '_' +
                   tag + '_evecs'] = isc_evecs

        return isc_diag


class JCDiagonal(AtomsProperty):

    """
    JCDiagonal

    Produces a dictionary of diagonalised J coupling tensors for atom pairs
    in the system. The J coupling for a pair of nuclei i and j is defined as:

    .. math::

        J_{ij} = 10^19\\frac{h\\gamma_i\\gamma_j}{4\\pi^2}K

    where the gammas represent the gyromagnetic ratios of the nuclei and K is
    the J coupling reduced tensor found in a .magres file, in 10^19.T^2.J^-1.


    | Parameters:
    |   sel_i (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling. By default is None
    |                                   (= all of them).
    |   sel_j (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling with the ones in sel_i. By
    |                                   default is None (= same as sel_i).
    |   tag (str): name of the J coupling component to return. Magres files
    |              usually contain isc, isc_spin, isc_fc, isc_orbital_p and
    |              isc_orbital_d. Default is isc.
    |   isotopes (dict): dictionary of specific isotopes to use, by element
    |                    symbol. If the isotope doesn't exist an error will
    |                    be raised.
    |   isotope_list (list): list of isotopes, atom-by-atom. To be used if
    |                        different atoms of the same element are supposed
    |                        to be of different isotopes. Where a 'None' is
    |                        present will fall back on the previous
    |                        definitions. Where an isotope is present it
    |                        overrides everything else.
    |   self_coupling (bool): if True, include coupling of a nucleus with
    |                         itself. Otherwise excluded. Default is False.
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present. Default is False.

    | Returns: 
    |   dip_dict (dict): Dictionary of couplings by atomic index pair.

    """

    default_name = 'jc_diagonal'
    default_params = {
        'sel_i': None,
        'sel_j': None,
        'tag': 'isc',
        'isotopes': {},
        'isotope_list': None,
        'self_coupling': False,
        'force_recalc': False,
    }

    @staticmethod
    @_has_isc_check
    def extract(s, sel_i, sel_j, tag, isotopes, isotope_list, self_coupling,
                force_recalc):

        # Compute the diagonalised eigenvectors if necessary
        iname_pairs = ISCDiagonal.default_name + '_' + tag + '_pairs'
        iname_evals = ISCDiagonal.default_name + '_' + tag + '_evals'
        iname_evecs = ISCDiagonal.default_name + '_' + tag + '_evecs'
        if iname_pairs not in s.info or force_recalc:
            iprop = ISCDiagonal(tag=tag)
            iprop(s)
        all_pairs = list(s.info[iname_pairs])
        all_evals = s.info[iname_evals]
        all_evecs = s.info[iname_evecs]

        # Selections
        if sel_i is None:
            sel_i = AtomSelection.all(s)
        elif not isinstance(sel_i, AtomSelection):
            sel_i = AtomSelection(s, sel_i)

        if sel_j is None:
            sel_j = sel_i
        elif not isinstance(sel_j, AtomSelection):
            sel_j = AtomSelection(s, sel_j)

        # Find gammas
        elems = s.get_chemical_symbols()
        _nmr_data = _get_nmr_data()

        gammas = _get_isotope_data(elems, 'gamma', isotopes, isotope_list)

        sel_pairs = [(i, j) for i in sel_i.indices
                     for j in sel_j.indices]
        if not self_coupling:
            sel_pairs = [p for p in sel_pairs if p[0] != p[1]]

        jc_dict = {}
        for sp in sel_pairs:
            try:
                i = all_pairs.index(sp)
            except ValueError:
                continue
            evals = _J_constant(all_evals[i], gammas[sp[0]], gammas[sp[1]])
            evecs = all_evecs[i]
            jc_dict[sp] = {'evals': evals, 'evecs': evecs}

        return jc_dict


class JCIsotropy(AtomsProperty):

    """
    JCIsotropy

    Produces a dictionary of J coupling isotropies for atom pairs
    in the system. See JCDiagonal for how reduced couplings are transformed
    into couplings.


    | Parameters:
    |   sel_i (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling. By default is None
    |                                   (= all of them).
    |   sel_j (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling with the ones in sel_i. By
    |                                   default is None (= same as sel_i).
    |   tag (str): name of the J coupling component to return. Magres files
    |              usually contain isc, isc_spin, isc_fc, isc_orbital_p and
    |              isc_orbital_d. Default is isc.
    |   isotopes (dict): dictionary of specific isotopes to use, by element
    |                    symbol. If the isotope doesn't exist an error will
    |                    be raised.
    |   isotope_list (list): list of isotopes, atom-by-atom. To be used if
    |                        different atoms of the same element are supposed
    |                        to be of different isotopes. Where a 'None' is
    |                        present will fall back on the previous
    |                        definitions. Where an isotope is present it
    |                        overrides everything else.
    |   self_coupling (bool): if True, include coupling of a nucleus with
    |                         itself. Otherwise excluded. Default is False.
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present. Default is False.

    | Returns: 
    |   dip_dict (dict): Dictionary of couplings by atomic index pair, in Hz.

    """

    default_name = 'jc_isotropy'
    default_params = {
        'sel_i': None,
        'sel_j': None,
        'tag': 'isc',
        'isotopes': {},
        'isotope_list': None,
        'self_coupling': False,
        'force_recalc': False,
    }

    @staticmethod
    @_has_isc_check
    def extract(s, sel_i, sel_j, tag, isotopes, isotope_list, self_coupling,
                force_recalc):

        jDiagProp = JCDiagonal(sel_i=sel_i, sel_j=sel_j, isotopes=isotopes,
                               tag=tag, isotope_list=isotope_list,
                               self_coupling=self_coupling,
                               force_recalc=force_recalc)
        jc_dict = jDiagProp(s)
        jc_keys = jc_dict.keys()
        jc_evals = [jc_dict[ij]['evals'] for ij in jc_keys]
        jc_iso = np.average(jc_evals, axis=1)

        return dict(zip(jc_dict.keys(), jc_iso))


class JCAnisotropy(AtomsProperty):

    """
    JCAnisotropy

    Produces a dictionary of J coupling anisotropies for atom pairs
    in the system. See JCDiagonal for how reduced couplings are transformed
    into couplings.


    | Parameters:
    |   sel_i (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling. By default is None
    |                                   (= all of them).
    |   sel_j (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling with the ones in sel_i. By
    |                                   default is None (= same as sel_i).
    |   tag (str): name of the J coupling component to return. Magres files
    |              usually contain isc, isc_spin, isc_fc, isc_orbital_p and
    |              isc_orbital_d. Default is isc.
    |   isotopes (dict): dictionary of specific isotopes to use, by element
    |                    symbol. If the isotope doesn't exist an error will
    |                    be raised.
    |   isotope_list (list): list of isotopes, atom-by-atom. To be used if
    |                        different atoms of the same element are supposed
    |                        to be of different isotopes. Where a 'None' is
    |                        present will fall back on the previous
    |                        definitions. Where an isotope is present it
    |                        overrides everything else.
    |   self_coupling (bool): if True, include coupling of a nucleus with
    |                         itself. Otherwise excluded. Default is False.
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present. Default is False.

    | Returns: 
    |   dip_dict (dict): Dictionary of couplings by atomic index pair, in Hz.

    """

    default_name = 'jc_anisotropy'
    default_params = {
        'sel_i': None,
        'sel_j': None,
        'tag': 'isc',
        'isotopes': {},
        'isotope_list': None,
        'self_coupling': False,
        'force_recalc': False,
    }

    @staticmethod
    @_has_isc_check
    def extract(s, sel_i, sel_j, tag, isotopes, isotope_list, self_coupling,
                force_recalc):

        jDiagProp = JCDiagonal(sel_i=sel_i, sel_j=sel_j, isotopes=isotopes,
                               tag=tag, isotope_list=isotope_list,
                               self_coupling=self_coupling,
                               force_recalc=force_recalc)
        jc_dict = jDiagProp(s)
        jc_keys = jc_dict.keys()
        jc_evals = [jc_dict[ij]['evals'] for ij in jc_keys]
        jc_evals = _haeb_sort(jc_evals)
        jc_aniso = _anisotropy(jc_evals)

        return dict(zip(jc_dict.keys(), jc_aniso))


class JCReducedAnisotropy(AtomsProperty):

    """
    JCReducedAnisotropy

    Produces a dictionary of J coupling reduced anisotropies for atom pairs
    in the system. See JCDiagonal for how reduced couplings are transformed
    into couplings.


    | Parameters:
    |   sel_i (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling. By default is None
    |                                   (= all of them).
    |   sel_j (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling with the ones in sel_i. By
    |                                   default is None (= same as sel_i).
    |   tag (str): name of the J coupling component to return. Magres files
    |              usually contain isc, isc_spin, isc_fc, isc_orbital_p and
    |              isc_orbital_d. Default is isc.
    |   isotopes (dict): dictionary of specific isotopes to use, by element
    |                    symbol. If the isotope doesn't exist an error will
    |                    be raised.
    |   isotope_list (list): list of isotopes, atom-by-atom. To be used if
    |                        different atoms of the same element are supposed
    |                        to be of different isotopes. Where a 'None' is
    |                        present will fall back on the previous
    |                        definitions. Where an isotope is present it
    |                        overrides everything else.
    |   self_coupling (bool): if True, include coupling of a nucleus with
    |                         itself. Otherwise excluded. Default is False.
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present. Default is False.

    | Returns: 
    |   dip_dict (dict): Dictionary of couplings by atomic index pair, in Hz.

    """

    default_name = 'jc_red_anisotropy'
    default_params = {
        'sel_i': None,
        'sel_j': None,
        'tag': 'isc',
        'isotopes': {},
        'isotope_list': None,
        'self_coupling': False,
        'force_recalc': False,
    }

    @staticmethod
    @_has_isc_check
    def extract(s, sel_i, sel_j, tag, isotopes, isotope_list, self_coupling,
                force_recalc):

        jDiagProp = JCDiagonal(sel_i=sel_i, sel_j=sel_j, isotopes=isotopes,
                               tag=tag, isotope_list=isotope_list,
                               self_coupling=self_coupling,
                               force_recalc=force_recalc)
        jc_dict = jDiagProp(s)
        jc_keys = jc_dict.keys()
        jc_evals = [jc_dict[ij]['evals'] for ij in jc_keys]
        jc_evals = _haeb_sort(jc_evals)
        jc_aniso = _anisotropy(jc_evals, reduced=True)

        return dict(zip(jc_dict.keys(), jc_aniso))


class JCAsymmetry(AtomsProperty):

    """
    JCAsymmetry

    Produces a dictionary of J coupling asymmetries for atom pairs
    in the system. See JCDiagonal for how reduced couplings are transformed
    into couplings.


    | Parameters:
    |   sel_i (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling. By default is None
    |                                   (= all of them).
    |   sel_j (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling with the ones in sel_i. By
    |                                   default is None (= same as sel_i).
    |   tag (str): name of the J coupling component to return. Magres files
    |              usually contain isc, isc_spin, isc_fc, isc_orbital_p and
    |              isc_orbital_d. Default is isc.
    |   isotopes (dict): dictionary of specific isotopes to use, by element
    |                    symbol. If the isotope doesn't exist an error will
    |                    be raised.
    |   isotope_list (list): list of isotopes, atom-by-atom. To be used if
    |                        different atoms of the same element are supposed
    |                        to be of different isotopes. Where a 'None' is
    |                        present will fall back on the previous
    |                        definitions. Where an isotope is present it
    |                        overrides everything else.
    |   self_coupling (bool): if True, include coupling of a nucleus with
    |                         itself. Otherwise excluded. Default is False.
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present. Default is False.

    | Returns: 
    |   dip_dict (dict): Dictionary of couplings by atomic index pair, in Hz.

    """

    default_name = 'jc_asymmetry'
    default_params = {
        'sel_i': None,
        'sel_j': None,
        'tag': 'isc',
        'isotopes': {},
        'isotope_list': None,
        'self_coupling': False,
        'force_recalc': False,
    }

    @staticmethod
    @_has_isc_check
    def extract(s, sel_i, sel_j, tag, isotopes, isotope_list, self_coupling,
                force_recalc):

        jDiagProp = JCDiagonal(sel_i=sel_i, sel_j=sel_j, isotopes=isotopes,
                               tag=tag, isotope_list=isotope_list,
                               self_coupling=self_coupling,
                               force_recalc=force_recalc)
        jc_dict = jDiagProp(s)
        jc_keys = jc_dict.keys()
        jc_evals = [jc_dict[ij]['evals'] for ij in jc_keys]
        jc_evals = _haeb_sort(jc_evals)
        jc_asymm = _asymmetry(jc_evals)

        return dict(zip(jc_dict.keys(), jc_asymm))


class JCSpan(AtomsProperty):

    """
    JCSpan

    Produces a dictionary of J coupling spans for atom pairs
    in the system. See JCDiagonal for how reduced couplings are transformed
    into couplings.


    | Parameters:
    |   sel_i (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling. By default is None
    |                                   (= all of them).
    |   sel_j (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling with the ones in sel_i. By
    |                                   default is None (= same as sel_i).
    |   tag (str): name of the J coupling component to return. Magres files
    |              usually contain isc, isc_spin, isc_fc, isc_orbital_p and
    |              isc_orbital_d. Default is isc.
    |   isotopes (dict): dictionary of specific isotopes to use, by element
    |                    symbol. If the isotope doesn't exist an error will
    |                    be raised.
    |   isotope_list (list): list of isotopes, atom-by-atom. To be used if
    |                        different atoms of the same element are supposed
    |                        to be of different isotopes. Where a 'None' is
    |                        present will fall back on the previous
    |                        definitions. Where an isotope is present it
    |                        overrides everything else.
    |   self_coupling (bool): if True, include coupling of a nucleus with
    |                         itself. Otherwise excluded. Default is False.
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present. Default is False.

    | Returns: 
    |   dip_dict (dict): Dictionary of couplings by atomic index pair, in Hz.

    """

    default_name = 'jc_span'
    default_params = {
        'sel_i': None,
        'sel_j': None,
        'tag': 'isc',
        'isotopes': {},
        'isotope_list': None,
        'self_coupling': False,
        'force_recalc': False,
    }

    @staticmethod
    @_has_isc_check
    def extract(s, sel_i, sel_j, tag, isotopes, isotope_list, self_coupling,
                force_recalc):

        jDiagProp = JCDiagonal(sel_i=sel_i, sel_j=sel_j, isotopes=isotopes,
                               tag=tag, isotope_list=isotope_list,
                               self_coupling=self_coupling,
                               force_recalc=force_recalc)
        jc_dict = jDiagProp(s)
        jc_keys = jc_dict.keys()
        jc_evals = [jc_dict[ij]['evals'] for ij in jc_keys]
        jc_evals = _haeb_sort(jc_evals)
        jc_span = _span(jc_evals)

        return dict(zip(jc_dict.keys(), jc_span))


class JCSkew(AtomsProperty):

    """
    JCSkew

    Produces a dictionary of J coupling skews for atom pairs
    in the system. See JCDiagonal for how reduced couplings are transformed
    into couplings.


    | Parameters:
    |   sel_i (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling. By default is None
    |                                   (= all of them).
    |   sel_j (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling with the ones in sel_i. By
    |                                   default is None (= same as sel_i).
    |   tag (str): name of the J coupling component to return. Magres files
    |              usually contain isc, isc_spin, isc_fc, isc_orbital_p and
    |              isc_orbital_d. Default is isc.
    |   isotopes (dict): dictionary of specific isotopes to use, by element
    |                    symbol. If the isotope doesn't exist an error will
    |                    be raised.
    |   isotope_list (list): list of isotopes, atom-by-atom. To be used if
    |                        different atoms of the same element are supposed
    |                        to be of different isotopes. Where a 'None' is
    |                        present will fall back on the previous
    |                        definitions. Where an isotope is present it
    |                        overrides everything else.
    |   self_coupling (bool): if True, include coupling of a nucleus with
    |                         itself. Otherwise excluded. Default is False.
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present. Default is False.

    | Returns: 
    |   dip_dict (dict): Dictionary of couplings by atomic index pair, in Hz.

    """

    default_name = 'jc_skew'
    default_params = {
        'sel_i': None,
        'sel_j': None,
        'tag': 'isc',
        'isotopes': {},
        'isotope_list': None,
        'self_coupling': False,
        'force_recalc': False,
    }

    @staticmethod
    @_has_isc_check
    def extract(s, sel_i, sel_j, tag, isotopes, isotope_list, self_coupling,
                force_recalc):

        jDiagProp = JCDiagonal(sel_i=sel_i, sel_j=sel_j, isotopes=isotopes,
                               tag=tag, isotope_list=isotope_list,
                               self_coupling=self_coupling,
                               force_recalc=force_recalc)
        jc_dict = jDiagProp(s)
        jc_keys = jc_dict.keys()
        jc_evals = [jc_dict[ij]['evals'] for ij in jc_keys]
        jc_evals = _haeb_sort(jc_evals)
        jc_skew = _asymmetry(jc_evals)

        return dict(zip(jc_dict.keys(), jc_skew))


class JCQuaternion(AtomsProperty):

    """
    JCQuaternion

    Produces a dictionary of J ase.Quaternion objects expressing the
    orientation of the J coupling tensors with respect to the cartesian axes
    for atom pairs in the system.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   sel_i (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling. By default is None
    |                                   (= all of them).
    |   sel_j (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to return the J
    |                                   coupling with the ones in sel_i. By
    |                                   default is None (= same as sel_i).
    |   tag (str): name of the J coupling component to return. Magres files
    |              usually contain isc, isc_spin, isc_fc, isc_orbital_p and
    |              isc_orbital_d. Default is isc.
    |   isotopes (dict): dictionary of specific isotopes to use, by element
    |                    symbol. If the isotope doesn't exist an error will
    |                    be raised.
    |   isotope_list (list): list of isotopes, atom-by-atom. To be used if
    |                        different atoms of the same element are supposed
    |                        to be of different isotopes. Where a 'None' is
    |                        present will fall back on the previous
    |                        definitions. Where an isotope is present it
    |                        overrides everything else.
    |   self_coupling (bool): if True, include coupling of a nucleus with
    |                         itself. Otherwise excluded. Default is False.
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present. Default is False.

    | Returns: 
    |   dip_dict (dict): Dictionary of couplings by atomic index pair, in Hz.

    """

    default_name = 'jc_skew'
    default_params = {
        'sel_i': None,
        'sel_j': None,
        'tag': 'isc',
        'isotopes': {},
        'isotope_list': None,
        'self_coupling': False,
        'force_recalc': False,
    }

    @staticmethod
    @_has_isc_check
    def extract(s, sel_i, sel_j, tag, isotopes, isotope_list, self_coupling,
                force_recalc):

        jDiagProp = JCDiagonal(sel_i=sel_i, sel_j=sel_j, isotopes=isotopes,
                               tag=tag, isotope_list=isotope_list,
                               self_coupling=self_coupling,
                               force_recalc=force_recalc)
        jc_dict = jDiagProp(s)
        jc_keys = jc_dict.keys()
        jc_evecs = [jc_dict[ij]['evecs'] for ij in jc_keys]
        jc_quat = _evecs_2_quat(jc_evecs)

        return dict(zip(jc_dict.keys(), jc_quat))



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Definition of CastepSubmitter class.

A basic "rolling" submitter for Castep calculations, grabbing from one folder
and depositing results in another.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import glob
import shutil
import tempfile
import numpy as np
import subprocess as sp

from soprano.utils import seedname, safe_communicate
from soprano.hpc.submitter import Submitter
from ase.calculators.castep import create_castep_keywords


class CastepSubmitter(Submitter):

    def set_parameters(self, folder_in, folder_out, castep_command,
                       castep_path=None,
                       copy_extensions=['.castep'],
                       pspot_files=[],
                       dryrun_test=False):
        """Set the parameters of the CASTEP Submitter

        | Args:
        |   folder_in (str): path of the folder to extract cell files from
        |   folder_out (str): path of the folder where the results will be
        |                     saved
        |   castep_command (str): command used to call the CASTEP executable
        |                         on this system
        |   castep_path (Optional[str]): folder where the CASTEP executable is
        |                                located (if not part of the system
        |                                PATH)
        |   pspot_files (Optional[list[str]]): additional pseudopotential
        |                                      files to be copied in the input
        |                                      temporary folders
        |   copy_extensions (Optional[list[str]]): extensions of output files
        |                                          to copy to the output
        |                                          folder (by default only
        |                                          .castep file)
        |   dryrun_test (Optional[bool]): run a dryrun test on files before
        |                                 actually running the calculation.
        |                                 Off by default.

        """

        # Check for existence of the folders, if not present create them
        try:
            os.mkdir(folder_in)
        except OSError:
            pass

        try:
            os.mkdir(folder_out)
        except OSError:
            pass

        self.folder_in = folder_in
        self.folder_out = folder_out
        self.castep_command = castep_command
        self.pspots = [os.path.abspath(pp) for pp in pspot_files]
        self.cp_ext = copy_extensions
        self.drun = dryrun_test

        self.castpath = castep_path

        # Finally add the castep_path to PATH if needed
        if castep_path is not None:
            sys.path.append(castep_path)

    def start_run(self):
        # Initialize the CASTEP keywords file in a dedicated temporary folder
        self.kwdir = tempfile.mkdtemp()
        self.log('Creating CASTEP keywords in folder '
                 '{0}\n'.format(self.kwdir))
        # Avoid the annoying print out to screen!
        _stdout, sys.stdout = sys.stdout, self._log
        create_castep_keywords(self.castep_command,
                               os.path.join(self.kwdir,
                                            'castep_keywords.py'))
        sys.stdout = _stdout
        sys.path.append(self.kwdir)
        self.log('CASTEP keywords created\n')

    def next_job(self):
        """Grab the next job from folder_in"""

        cfile_list = glob.glob(os.path.join(self.folder_in, '*.cell'))
        if len(cfile_list) == 0:
            return None

        cfile = cfile_list[0]
        name = seedname(cfile)
        self.log('Starting job {0}\n'.format(name))
        files = [cfile]
        # Check if .param file is available too
        if os.path.isfile(os.path.join(self.folder_in, name + '.param')):
            files += [os.path.join(self.folder_in, name + '.param')]
        job = {
            'name': name,
            'args': {
                'files': files
            }
        }

        return job

    def setup_job(self, name, args, folder):
        """Copy files to temporary folder to prepare for execution"""

        for f in args['files']:
            shutil.move(f, folder)
        # Also copy the pseudopotentials
        for pp in self.pspots:
            shutil.copy(pp, folder)

        success = True

        self.log('Copying files for job {0}\n'.format(name))

        # Perform dryrun test if required
        if self.drun:
            self.log('Performing DRYRUN\n')
            dry_proc = sp.Popen([self.castep_command, name, '--dryrun'],
                                cwd=folder, stdout=sp.PIPE,
                                stderr=sp.PIPE)
            stdout, stderr = safe_communicate(dry_proc)
            # When it's finished...
            try:
                castfile = open(os.path.join(folder, name + '.castep'))
            except IOError:
                return False
            # Does the file contain the required lines?
            drline1 = "|       DRYRUN finished ...                       |"
            drline2 = "|       No problems found with input files.       |"

            castlines = castfile.readlines()
            success = False
            for i, l in enumerate(castlines):
                if drline1 in l:
                    if drline2 in castlines[i+1]:
                        success = True
                        break

            # Ok, now remove the CASTEP file
            castfile.close()
            os.remove(castfile.name)

        return success

    def finish_job(self, name, args, folder):
        """Save required output files to the output folder"""

        for cext in self.cp_ext:
            files = glob.glob(os.path.join(folder, '*'+cext))
            for f in files:
                shutil.move(f, os.path.join(self.folder_out,
                                            os.path.basename(f)))

    def finish_run(self):
        """Try removing the temporary keywords directory"""
        try:
            shutil.rmtree(self.kwdir)
        except OSError:
            self.log("Could not delete temporary castep_keywords.py"
                     "directory at {0}".format(self.kwdir))
        sys.path.remove(self.kwdir)

        if self.castpath is not None:
            sys.path.remove(castep_path)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
powder.py

Generic powder averaging scheme class for inheritance
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

class PowderScheme(object):

	def __init__(self, mode):

		if mode not in ('sphere', 'hemisphere', 'octant'):
			raise ValueError('Invalid mode passed to powder averaging scheme')

		self.mode = mode

	def get_orient_angles(self, N):
		pass

	def get_orient_trig(self, N):
		pass

	def get_orient_points(self, N):
		pass


# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Contains classes, modules and functions relevant to Properties,
a catch-all term for things we might want to extract or calculate from
Atoms and AtomsCollections. Some will require running an external ASE
calculator first, some will just work on their own, some will require
some calculations and parameters.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import

from soprano.properties.atomsproperty import AtomsProperty



# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
measure.py

Contains utility functions for measuring distances and other quantities at
the level of a single ase.Atoms object.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np

from soprano.properties.linkage import Bonds
from soprano.utils import (minimum_periodic, get_bonding_graph,
                           get_bonding_distance)


def euclideanDistance(s, i, j, periodic=True):
    """
    Return the distance between two atoms, in Angstroms.

    | Parameters:
    |   s (ase.Atoms): the structure on which to compute the distance
    |   i (int): index of the first atom
    |   j (int): index of the second atom
    |   periodic (bool): whether to account for periodic boundaries when 
    |                    computing the distance; default is True

    | Returns:
    |   r (float): computed distance
    """

    if i == j:
        return 0.0

    cell = s.get_cell()
    pos = s.get_positions()

    r = pos[j]-pos[i]

    if periodic:
        r, _ = minimum_periodic(r[None, :], cell)

    return np.linalg.norm(r)


def bondDistance(s, i, j, bond_matrix=None):
    """
    Return the distance in number of bonds between two atoms. Returns -1 if
    the two atoms are not connected. Requires NetworkX to be installed to
    work.

    | Parameters:
    |   s (ase.Atoms): the structure on which to compute the distance
    |   i (int): index of the first atom
    |   j (int): index of the second atom
    |   bond_matrix (np.ndarray): pre-computed bond matrix to use. If not
    |                             provided, will be calculated with default
    |                             parameters

    | Returns:
    |   r (int): computed bond distance
    """

    if bond_matrix is None:
        bprop = Bonds(return_matrix=True)
        _, bond_matrix = bprop(s)

    graph = get_bonding_graph(bond_matrix)
    r = get_bonding_distance(graph, i, j)

    return r



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""2D mapping algorithms"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from scipy.cluster import vq
from scipy import linalg


def _check_dimensionality(mapfunc):
    def mapfunc_decorated(p):
        if p.shape[0] < 2:
            raise ValueError('Dimensionality of the problem'
                             'too small for 2D mapping')
        return mapfunc(p)

    return mapfunc_decorated


@_check_dimensionality
def total_principal_component(p):
    """Total principal component mapping"""

    covmat = np.cov(p.T)
    evals, evecs = np.linalg.eig(covmat)
    A = evecs[:, :2]
    return np.tensordot(A, p, axes=(0, 1))


@_check_dimensionality
def classcond_principal_component(p):
    """Class conditional principal component mapping (Clafic)"""

    # Use k-means clustering to create two classes
    centroids, dist = vq.kmeans(p, 2)
    clusts, cdists = vq.vq(p, centroids)

    # Split the two clusters
    p1 = p[np.where(clusts == 0)]
    p2 = p[np.where(clusts == 1)]

    S1 = np.cov(p1.T)
    S2 = np.cov(p2.T)

    evals1, evecs1 = linalg.eig(a=S1)
    evals2, evecs2 = linalg.eig(a=S2)
    A = np.concatenate((evecs1[:, 0][:, None], evecs2[:, 0][:, None]), axis=1)
    return np.tensordot(A, p, axes=(0, 1))


@_check_dimensionality
def standard_classcond_component(p):
    """Standardized class conditional principal component mapping (Fukunaga-
    Koontz)"""

    # Use k-means clustering to create two classes
    centroids, dist = vq.kmeans(p, 2)
    clusts, cdists = vq.vq(p, centroids)

    # Split the two clusters
    p1 = p[np.where(clusts == 0)]
    p2 = p[np.where(clusts == 1)]

    S1 = np.cov(p1.T)
    S2 = np.cov(p2.T)

    evals, evecs = linalg.eig(a=S1, b=(S1+S2))
    A = evecs[:, :2]
    return np.tensordot(A, p, axes=(0, 1))


@_check_dimensionality
def optimal_discriminant_plane(p):
    """Optimal discriminant plane mapping (using Fischer direction)"""

    # First, split in two classes
    centroids, dist = vq.kmeans(p, 2)
    clusts, cdists = vq.vq(p, centroids)

    p1 = p[np.where(clusts == 0)]
    p2 = p[np.where(clusts == 1)]

    # Find the average points
    m1 = np.average(p1, axis=0)
    m2 = np.average(p2, axis=0)
    mdiff = m1-m2

    B = np.dot(mdiff[:, None], mdiff[None, :])

    S1 = np.cov(p1.T)
    S2 = np.cov(p2.T)
    S = S1+S2

    # So the fisher direction is found analytically...
    A = np.zeros((B.shape[0], 2))
    invS = np.linalg.inv(S)
    A[:, 0] = np.dot(invS, mdiff)
    A[:, 0] /= np.linalg.norm(A[:, 0])  # Normalized
    # And then the second one
    z = np.dot(invS, A[:, 0])
    A[:, 1] = A[:, 0]-np.dot(A[:, 0], z)*z

    return np.tensordot(A, p, axes=(0, 1))



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes and functions for interfacing with the SIMPSON spin dynamics software.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import numpy as np
from collections import namedtuple
from soprano.properties.nmr import (MSIsotropy, MSReducedAnisotropy,
                                    MSAsymmetry,
                                    MSQuaternion,
                                    EFGQuadrupolarConstant, EFGAsymmetry,
                                    EFGQuaternion,
                                    DipolarCoupling)
from soprano.selection import AtomSelection
from soprano.data.nmr import _get_nmr_data, _el_iso

_spinsys_template = """spinsys {{
{header}
{ms}
{efg}
{dipolar}
}}
"""

_header_template = """
channels {channels}
nuclei {nuclei}
"""


def write_spinsys(s, isotope_list=None, use_ms=False, ms_iso=False,
                  q_order=0, dip_sel=None, path=None):
    """
    Write a .spinsys input file for use with SIMPSON, given the details of a
    system. This is meant to be a low-level function, used by other higher
    level interfaces in NMRCalculator.

    | Args:
    |   s (ase.Atoms): atomic structure containing the desired spins. All
    |                  atoms will be included - if that is not the desired
    |                  result, this should be accomplished by making this a
    |                  subset of the full structure.
    |   isotope_list ([int]): list of isotopes for each element in the system.
    |                         If left to None, default NMR-active isotopes
    |                         will be used.
    |   use_ms (bool): if True, include shift interactions from magnetic
    |                  shieldings.
    |   ms_iso (bool): if True, all magnetic shieldings will be made
    |                  isotropic.
    |   q_order(int): if greater than 0, include quadrupolar interactions from
    |                   Electric Field Gradients at the given order (1 or 2).
    |   dip_sel (AtomSelection): if not None, include dipolar couplings
    |                            between atoms belonging to this set.
    |   path (str): path to save the newly created file to. If not provided,
    |               the contents will be simply returned as a string.

    | Returns:
    |   file_contents (str): spinsys file in string format. Only returned if
    |                        no save path is provided.

    """

    # Start by creating a proper isotope_list
    nmr_data = _get_nmr_data()

    nuclei = s.get_chemical_symbols()

    if isotope_list is None:
        isotope_list = [int(nmr_data[n]["iso"]) for n in nuclei]

    nuclei = [str(i)+n for i, n in zip(isotope_list, nuclei)]

    # Build header
    header = _header_template.format(channels=' '.join(set(nuclei)),
                                     nuclei=' '.join(nuclei))

    # Build MS block
    ms_block = ''

    if use_ms:

        msiso = MSIsotropy.get(s)
        if not ms_iso:
            msaniso = MSReducedAnisotropy.get(s)
            msasymm = MSAsymmetry.get(s)
            eulangs = np.array([q.euler_angles()
                                for q in MSQuaternion.get(s)])*180/np.pi
        else:
            msaniso = np.zeros(len(s))
            msasymm = np.zeros(len(s))
            eulangs = np.zeros((len(s), 3))

        for i, ms in enumerate(msiso):
            ms_block += ('shift {0} {1}p {2}p '
                         '{3} {4} {5} {6}\n').format(i+1,
                                                     ms, msaniso[i],
                                                     msasymm[i], *eulangs[i])

    # Build EFG block
    efg_block = ''

    if q_order > 0:
        if q_order > 2:
            raise ValueError('Invalid quadrupolar order')
        Cq = EFGQuadrupolarConstant(isotope_list=isotope_list)(s)
        eta_q = EFGAsymmetry.get(s)
        eulangs = np.array([q.euler_angles()
                            for q in EFGQuaternion.get(s)])*180/np.pi
        for i, cq in enumerate(Cq):
            if cq == 0:
                continue
            efg_block += ('quadrupole {0} {1} {2} {3}'
                          ' {4} {5} {6}\n').format(i+1, q_order, cq, eta_q[i],
                                                   *eulangs[i])

    # Build dipolar block
    dip_block = ''

    if dip_sel is not None and len(dip_sel) > 1:
        dip = DipolarCoupling(sel_i=dip_sel, isotope_list=isotope_list)(s)
        for (i, j), (d, v) in dip.items():
            a, b = (np.array([np.arccos(-v[2]),
                              np.arctan2(-v[1],
                                         -v[0])]) % (2*np.pi)) * 180/np.pi
            dip_block += ('dipole {0} {1} {2} {3}'
                          ' {4} 0.0\n').format(i+1, j+1, d*2*np.pi, a, b)

    out_file = _spinsys_template.format(header=header, ms=ms_block,
                                        efg=efg_block, dipolar=dip_block)

    if path is None:
        return out_file
    else:
        with open(path, 'w') as of:
            of.write(out_file)


def load_simpson_dat(filename):
    """Load a SIMPSON output .dat file and return it as a numpy array."""

    dat = np.loadtxt(filename)
    return np.concatenate([dat[:, 0, None],
                           (dat[:, 1] + 1.0j*dat[:, 2])[:, None]], axis=1)


SimpsonTemplates = namedtuple('SimpsonTemplates',
                              """BASIC_SPECTRUM,
                              BROADENED_SPECTRUM
                              """)

SimpsonTemplates = SimpsonTemplates(
    BASIC_SPECTRUM={
        'pars': {},
        'pulseq': """
global par
delay [expr 1e6/$par(sw)]
store 1
acq $par(np) 1
""",
        'main': """
global par

set f [fsimpson]
fsave $f $par(name)_fid.dat -xreim
fft $f
fsave $f $par(name)_spe.dat -xreim
puts "Simulation complete"
"""
    },

    BROADENED_SPECTRUM={
        'pars': {
            'broad': '1',
        },
        'pulseq': """
global par
delay [expr 1e6/$par(sw)]
store 1
acq $par(np) 1
""",
        'main': """
global par

set f [fsimpson]
fsave $f $par(name)_fid.dat -xreim
faddlb $f $par(broad) 0
fft $f
fsave $f $par(name)_spe.dat -xreim
puts "Simulation complete"
"""
    })


def _check_template_pars(f):
    def decorated_f(self, *args, **kwargs):
        f(self, *args, **kwargs)

        # Scan for missing parameters
        par_re = re.compile('\$par\(([0-9a-zA-Z_]+)\)')

        req_pars = set(par_re.findall(self.main) +
                       par_re.findall(self.pulseq))
        # Remove 'name' since that's defined by default
        req_pars = req_pars.difference(['name'])

        if not req_pars.issubset(set(self.pars.keys())):
            print('WARNING: some parameters required for this sequence seem'
                  ' to be not defined')

    return decorated_f


class SimpsonSequence:

    """SimpsonSequence

    A class storing parameters and scripts for the production of a SIMPSON
    input file. The parameters of the simulation are stored in a dictionary
    member accessible as .pars and can be set at will.

    | Args:
    |   spinsys_source (str): path of the .spinsys file to use in the
    |                         simulation

    """

    def __init__(self, spinsys_source):

        self.spinsys_source = spinsys_source

        self.pars = {
            'proton_frequency': '4e7',
            'start_operator': 'Inx',
            'detect_operator': 'Inx',
            'np': '8192',
            'sw': '8000',
            'num_cores': '1',
            'crystal_file': 'alpha0beta0'
        }

        self.apply_template(SimpsonTemplates.BASIC_SPECTRUM)

    @_check_template_pars
    def apply_template(self, template):
        """Apply an existing sequence template from SimpsonTemplates, 
        including default parameters.

        | Args:
        |   template (dict): the template to apply

        """

        self.pars.update(template['pars'])
        self.pulseq = template['pulseq']
        self.main = template['main']

    @_check_template_pars
    def apply_custom_template(self, pars, pulseq, main):
        """Apply a custom sequence template, defined by pars, pulse sequence
        and main script.

        | Args:
        |   pars (dict): default parameters for this sequence. Use an empty
        |                dict if none of relevance.
        |   pulseq (str): the script for the pulse sequence block.
        |   main (str): the script for the main block.

        """

        self.pars.update(pars)
        self.pulseq = pulseq
        self.main = main

    @_check_template_pars
    def set_parameters(self, **new_pars):
        """Set one or more parameters of the calculation. Compared to editing
        .pars directly it applies some more checks and is safer. Pass the new
        parameters as named arguments to this function.

        """

        self.pars.update(new_pars)

    def write_input(self, path=None):
        """Print out the .in file.

        | Args:
        |   path (str): path to save the newly created file to. If not provided,
        |               the contents will be simply returned as a string.

        """

        outf = 'source {0}\n\n'.format(self.spinsys_source)

        # Write out par block
        outf += 'par {\n'
        for p, val in self.pars.items():
            outf += '\t{0} {1}\n'.format(p, val)
        outf += '}\n\n'

        # Write out pulseq
        outf += 'proc pulseq {} {\n'
        outf += self.pulseq
        outf += '\n}\n\n'

        # Write out main
        outf += 'proc main {} {\n'
        outf += self.main
        outf += '\n}\n\n'

        if path is None:
            return outf
        else:
            with open(path, 'w') as of:
                of.write(outf)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Definition of AtomsProperty class.

A generic template class that specific Properties will inherit from.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from soprano.collection import AtomsCollection


class AtomsProperty(object):

    default_name = 'generic_property'
    default_params = {}

    def __init__(self, name=None, **params):
        """Initialize an AtomsProperty and set its parameters.
        The AtomsProperty instance can then be called with a structure as its
        only argument to get the property with the given parameters.

        | Args:
        |   name (str): a name to give to this specific instance of the
        |               property (will be used to store it as array if
        |               requested)
        |   params: named arguments specific to this type of property

        """

        if name is not None:
            self.name = name
        else:
            self.name = self.default_name
        # Validate the passed parameters
        self.params = dict(self.default_params)
        for p in params:
            if p not in self.params:
                raise ValueError('Invalid argument passed to'
                                 ' \'{0}\''.format(self.__class__.__name__))
            else:
                self.params[p] = params[p]

    @classmethod
    def get(self, s, store_array=False, **kwargs):
        """Extract the given property using the default parameters
        on an Atoms object s

        | Args:
        |   s (ase.Atoms or AtomsCollection): the structure or collection
        |                                     from which to extract the
        |                                     property
        |   store_array (bool): if s is a collection, whether to store the
        |                       resulting data as an array in the collection
        |                       using the default name for this property
        |

        | Returns:
        |   property: the value of the property for the given structure or
        |             a list of values if a collection has been passed
        |

        """

        if isinstance(s, AtomsCollection):
            arr = s.all.map(self.get, **kwargs)
            if store_array:
                s.set_array(self.default_name, arr)
            return arr
        else:
            params = dict(self.default_params)
            params.update(kwargs)
            return self.extract(s, **params)

    @staticmethod
    def extract(s, **params):
        """Extract the given property with given parameters from an Atoms
        object.

        | Args:
        |   s (ase.Atoms): the structure from which to extract the property
        |   params: named arguments specific to this type of property
        |

        | Returns:
        |   property: the value of the property for the given structure and
        |             parameters
        |

        """

        # Do something specific to get the property of interest
        # Then return the value

        pass

    def __call__(self, s, store_array=False):
        """Calling the AtomsProperty returns the value of the property as
        extracted with the parameters of this specific instance.

        | Args:
        |   s (ase.Atoms or AtomsCollection): the structure or collection
        |                                     from which to extract the
        |                                     property
        |   store_array (bool): if s is a collection, whether to store the
        |                       resulting data as an array in the collection
        |                       using the given name for this instance
        |

        | Returns:
        |   property: the value of the property for the given structure or
        |             a list of values if a collection has been passed
        |

        """

        if isinstance(s, AtomsCollection):
            arr = s.all.map(self.__call__)
            if store_array:
                s.set_array(self.name, arr)
            return arr
        else:
            return self.extract(s, **self.params)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of AtomsProperties that relate to NMR shieldings/shifts"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from soprano.properties import AtomsProperty
from soprano.nmr.utils import (_haeb_sort, _anisotropy, _asymmetry,
                               _span, _skew, _evecs_2_quat)


def _has_ms_check(f):
    # Decorator to add a check for the magnetic shieldings array
    def decorated_f(s, *args, **kwargs):
        if not (s.has('ms')):
            raise RuntimeError('No magnetic shielding data found for this'
                               ' system')
        return f(s, *args, **kwargs)

    return decorated_f


class MSDiagonal(AtomsProperty):

    """
    MSDiagonal

    Produces an array containing eigenvalues and eigenvectors for the
    symmetric part of each magnetic shielding tensor in the system. By default
    saves them as part of the Atoms' arrays as well.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   save_array (bool): if True, save the diagonalised tensors in the
    |                      Atoms object as an array. By default True.

    | Returns:
    |   ms_diag (np.ndarray): list of eigenvalues and eigenvectors

    """

    default_name = 'ms_diagonal'
    default_params = {
        'save_array': True
    }

    @staticmethod
    @_has_ms_check
    def extract(s, save_array):

        ms_diag = [np.linalg.eigh((ms+ms.T)/2.0) for ms in s.get_array('ms')]
        ms_evals, ms_evecs = [np.array(a) for a in zip(*ms_diag)]

        if save_array:
            s.set_array(MSDiagonal.default_name + '_evals', ms_evals)
            # Store also the Haeberlen sorted version
            s.set_array(MSDiagonal.default_name +
                        '_evals_hsort', _haeb_sort(ms_evals))
            s.set_array(MSDiagonal.default_name + '_evecs', ms_evecs)

        return np.array([dict(zip(('evals', 'evecs'), ms)) for ms in ms_diag])


class MSIsotropy(AtomsProperty):

    """
    MSIsotropy

    Produces an array containing the magnetic shielding isotropies in a system
    (ppm, with reference if provided).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   ref (float): reference frequency. If provided, the chemical shift
    |                will be returned instead of the magnetic shielding.

    | Returns:
    |   ms_iso (np.ndarray): list of shieldings/shifts

    """

    default_name = 'ms_isotropy'
    default_params = {
        'ref': None
    }

    @staticmethod
    @_has_ms_check
    def extract(s, ref):

        ms_iso = np.trace(s.get_array('ms'), axis1=1, axis2=2)/3.0

        # Referenced?
        if ref is not None:
            ms_iso = ref - ms_iso

        return ms_iso


class MSAnisotropy(AtomsProperty):

    """
    MSAnisotropy

    Produces an array containing the magnetic shielding anisotropies in a
    system (ppm).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.

    | Returns:
    |   ms_list (np.ndarray): list of anisotropies

    """

    default_name = 'ms_anisotropy'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    @_has_ms_check
    def extract(s, force_recalc):

        if (not s.has(MSDiagonal.default_name + '_evals_hsort') or
                force_recalc):
            MSDiagonal.get(s)

        ms_evals = s.get_array(MSDiagonal.default_name + '_evals_hsort')

        return _anisotropy(ms_evals)


class MSReducedAnisotropy(AtomsProperty):

    """
    MSReducedAnisotropy

    Produces an array containing the magnetic shielding reduced anisotropies
    in a system (ppm).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.

    | Returns:
    |   ms_list (np.ndarray): list of reduced anisotropies

    """

    default_name = 'ms_red_anisotropy'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    @_has_ms_check
    def extract(s, force_recalc):

        if (not s.has(MSDiagonal.default_name + '_evals_hsort') or
                force_recalc):
            MSDiagonal.get(s)

        ms_evals = s.get_array(MSDiagonal.default_name + '_evals_hsort')

        return _anisotropy(ms_evals, reduced=True)


class MSAsymmetry(AtomsProperty):

    """
    MSAsymmetry

    Produces an array containing the magnetic shielding asymmetries
    in a system (adimensional).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.

    | Returns:
    |   ms_list (np.ndarray): list of asymmetries

    """

    default_name = 'ms_asymmetry'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    @_has_ms_check
    def extract(s, force_recalc):

        if (not s.has(MSDiagonal.default_name + '_evals_hsort') or
                force_recalc):
            MSDiagonal.get(s)

        ms_evals = s.get_array(MSDiagonal.default_name + '_evals_hsort')

        return _asymmetry(ms_evals)


class MSSpan(AtomsProperty):

    """
    MSSpan

    Produces an array containing the magnetic shielding tensor span
    in a system (ppm).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.

    | Returns:
    |   ms_list (np.ndarray): list of spans

    """

    default_name = 'ms_span'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    @_has_ms_check
    def extract(s, force_recalc):

        if (not s.has(MSDiagonal.default_name + '_evals_hsort') or
                force_recalc):
            MSDiagonal.get(s)

        ms_evals = s.get_array(MSDiagonal.default_name + '_evals_hsort')

        return _span(ms_evals)


class MSSkew(AtomsProperty):

    """
    MSSkew

    Produces an array containing the magnetic shielding tensor skew
    in a system.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.

    | Returns:
    |   ms_list (np.ndarray): list of skews

    """

    default_name = 'ms_skew'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    @_has_ms_check
    def extract(s, force_recalc):

        if (not s.has(MSDiagonal.default_name + '_evals_hsort') or
                force_recalc):
            MSDiagonal.get(s)

        ms_evals = s.get_array(MSDiagonal.default_name + '_evals_hsort')

        return _skew(ms_evals)


class MSQuaternion(AtomsProperty):

    """
    MSQuaternion

    Produces a list of ase.Quaternion objects expressing the orientation of
    the MS tensors with respect to the cartesian axes.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.

    | Returns:
    |   ms_quat (np.ndarray): list of quaternions

    """

    default_name = 'ms_quats'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    @_has_ms_check
    def extract(s, force_recalc):

        if (not s.has(MSDiagonal.default_name + '_evecs') or
                force_recalc):
            MSDiagonal.get(s)

        ms_evecs = s.get_array(MSDiagonal.default_name + '_evecs')

        return _evecs_2_quat(ms_evecs)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of AtomsProperties that relate to NMR electric field
gradients"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from scipy import constants as cnst
from soprano.properties import AtomsProperty
from soprano.nmr.utils import (_haeb_sort, _anisotropy, _asymmetry,
                               _span, _skew, _evecs_2_quat)
from soprano.data.nmr import _get_nmr_data, _get_isotope_data, EFG_TO_CHI


def _has_efg_check(f):
    # Decorator to add a check for the electric field gradient array
    def decorated_f(s, *args, **kwargs):
        if not (s.has('efg')):
            raise RuntimeError('No electric field gradient data found for'
                               ' this system')
        return f(s, *args, **kwargs)

    return decorated_f


class EFGDiagonal(AtomsProperty):

    """
    EFGDiagonal

    Produces an array containing eigenvalues and eigenvectors for the
    symmetric part of each EFG tensor in the system. By default
    saves them as part of the Atoms' arrays as well.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   save_array (bool): if True, save the diagonalised tensors in the
    |                      Atoms object as an array. By default True.

    | Returns:
    |   efg_diag (np.ndarray): list of eigenvalues and eigenvectors

    """

    default_name = 'efg_diagonal'
    default_params = {
        'save_array': True
    }

    @staticmethod
    @_has_efg_check
    def extract(s, save_array):

        efg_diag = [np.linalg.eigh((efg+efg.T)/2.0)
                    for efg in s.get_array('efg')]
        efg_evals, efg_evecs = [np.array(a) for a in zip(*efg_diag)]

        if save_array:
            s.set_array(EFGDiagonal.default_name + '_evals', efg_evals)
            # Store also the Haeberlen sorted version
            s.set_array(EFGDiagonal.default_name +
                        '_evals_hsort', _haeb_sort(efg_evals))
            s.set_array(EFGDiagonal.default_name + '_evecs', efg_evecs)

        return np.array([dict(zip(('evals', 'evecs'), efg))
                         for efg in efg_diag])


class EFGVzz(AtomsProperty):

    """
    EFGVzz

    Produces an array containing the major component of the electric field
    gradient in a system (au).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.

    | Returns:
    |   efg_list (np.ndarray): list of Vzz values
    """

    default_name = 'efg_vzz'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc):

        if (not s.has(EFGDiagonal.default_name + '_evals_hsort') or
                force_recalc):
            EFGDiagonal.get(s)

        efg_evals = s.get_array(EFGDiagonal.default_name + '_evals_hsort')

        return efg_evals[:, -1]


class EFGAnisotropy(AtomsProperty):

    """
    EFGAnisotropy

    Produces an array containing the electric field gradient anisotropies in a
    system (au).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.

    | Returns:
    |   efg_list (np.ndarray): list of anisotropies

    """

    default_name = 'efg_anisotropy'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc):

        if (not s.has(EFGDiagonal.default_name + '_evals_hsort') or
                force_recalc):
            EFGDiagonal.get(s)

        efg_evals = s.get_array(EFGDiagonal.default_name + '_evals_hsort')

        return _anisotropy(efg_evals)


class EFGReducedAnisotropy(AtomsProperty):

    """
    EFGReducedAnisotropy

    Produces an array containing the electric field gradient reduced
    anisotropies in a system (au).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.

    | Returns:
    |   efg_list (np.ndarray): list of reduced anisotropies

    """

    default_name = 'efg_red_anisotropy'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc):

        if (not s.has(EFGDiagonal.default_name + '_evals_hsort') or
                force_recalc):
            EFGDiagonal.get(s)

        efg_evals = s.get_array(EFGDiagonal.default_name + '_evals_hsort')

        return _anisotropy(efg_evals, reduced=True)


class EFGAsymmetry(AtomsProperty):

    """
    EFGAsymmetry

    Produces an array containing the electric field gradient asymmetries
    in a system (adimensional).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.

    | Returns:
    |   efg_list (np.ndarray): list of asymmetries

    """

    default_name = 'efg_asymmetry'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc):

        if (not s.has(EFGDiagonal.default_name + '_evals_hsort') or
                force_recalc):
            EFGDiagonal.get(s)

        efg_evals = s.get_array(EFGDiagonal.default_name + '_evals_hsort')

        return _asymmetry(efg_evals)


class EFGSpan(AtomsProperty):

    """
    EFGSpan

    Produces an array containing the electric field gradient tensor span
    in a system (au).
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.

    | Returns:
    |   efg_list (np.ndarray): list of spans

    """

    default_name = 'efg_span'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc):

        if (not s.has(EFGDiagonal.default_name + '_evals_hsort') or
                force_recalc):
            EFGDiagonal.get(s)

        efg_evals = s.get_array(EFGDiagonal.default_name + '_evals_hsort')

        return _span(efg_evals)


class EFGSkew(AtomsProperty):

    """
    EFGSkew

    Produces an array containing the electric field gradient tensor skew
    in a system.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.

    | Returns:
    |   efg_list (np.ndarray): list of skews

    """

    default_name = 'efg_skew'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc):

        if (not s.has(EFGDiagonal.default_name + '_evals_hsort') or
                force_recalc):
            EFGDiagonal.get(s)

        efg_evals = s.get_array(EFGDiagonal.default_name + '_evals_hsort')

        return _skew(efg_evals)


class EFGQuadrupolarConstant(AtomsProperty):

    """
    EFGQuadrupolarConstant

    Produces an array containing the quadrupolar constant in Hz for every atom
    in a system. The constant will be zero for non-quadrupole active nuclei.
    Unless specified otherwise, the quadrupole moment of the most common
    NMR-active isotope is used.
    For reference: the value returned by this property is defined as

    .. math::

        \\frac{e^2qQ}{h}

    in Hz. It is important to keep in mind that therefore this represents a
    *frequency*; the corresponding 'omega' (pulsation) would be the same value
    multiplied by 2*pi. This is, for example, exactly the value required as
    input in Simpson's SPINSYS section.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.
    |   use_q_isotopes (bool): if True, always use the most common quadrupole
    |                          active isotope for each element, if there is
    |                          one.
    |   isotopes (dict): dictionary of specific isotopes to use, by element
    |                    symbol. If the isotope doesn't exist an error will
    |                    be raised. Overrides use_q_isotopes.
    |   isotope_list (list): list of isotopes, atom-by-atom. To be used if
    |                        different atoms of the same element are supposed
    |                        to be of different isotopes. Where a 'None' is
    |                        present will fall back on the previous
    |                        definitions. Where an isotope is present it
    |                        overrides everything else.

    | Returns:
    |   q_list (np.ndarray): list of quadrupole constants in Hz

    """

    default_name = 'efg_qconst'
    default_params = {
        'force_recalc': False,
        'use_q_isotopes': False,
        'isotopes': {},
        'isotope_list': None
    }

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc, use_q_isotopes, isotopes, isotope_list):

        if (not s.has(EFGDiagonal.default_name + '_evals_hsort') or
                force_recalc):
            EFGDiagonal.get(s)

        # First thing, build the isotope dictionary
        elems = s.get_chemical_symbols()

        # Is isotope list valid?
        if isotope_list is not None and len(isotope_list) != len(elems):
            print('WARNING - invalid isotope_list, ignoring')
            isotope_list = None

        q_list = _get_isotope_data(elems, 'Q', isotopes, isotope_list,
                                   use_q_isotopes)

        return EFG_TO_CHI*q_list*EFGVzz.get(s)


class EFGQuadrupolarProduct(AtomsProperty):

    """
    EFGQuadrupolarProduct

    Produces an array containing the quadrupolar product values
    in a system.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    .. math::

        \\P_Q = C_Q (1+frac{\eta_Q^2}{3})

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.
    |   use_q_isotopes (bool): if True, always use the most common quadrupole
    |                          active isotope for each element, if there is
    |                          one.
    |   isotopes (dict): dictionary of specific isotopes to use, by element
    |                    symbol. If the isotope doesn't exist an error will
    |                    be raised. Overrides use_q_isotopes.
    |   isotope_list (list): list of isotopes, atom-by-atom. To be used if
    |                        different atoms of the same element are supposed
    |                        to be of different isotopes. Where a 'None' is
    |                        present will fall back on the previous
    |                        definitions. Where an isotope is present it
    |                        overrides everything else.

    | Returns:
    |   Pq_list (np.ndarray): list of quadrupole products in Hz (units of Cq)

    """

    default_name = 'efg_qprod'
    default_params = {
        'force_recalc': False,
        'use_q_isotopes': False,
        'isotopes': {},
        'isotope_list': None
    }

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc, use_q_isotopes, isotopes, isotope_list):

        if (not s.has(EFGDiagonal.default_name + '_evals_hsort') or
                force_recalc):
            EFGDiagonal.get(s)

        # First thing, build the isotope dictionary
        elems = s.get_chemical_symbols()

        # Is isotope list valid?
        if isotope_list is not None and len(isotope_list) != len(elems):
            print('WARNING - invalid isotope_list, ignoring')
            isotope_list = None

        q_list = _get_isotope_data(elems, 'Q', isotopes, isotope_list,
                                   use_q_isotopes)

        return EFG_TO_CHI*q_list*EFGVzz.get(s) * (1+(EFGAsymmetry.get(s)**2)/3)**0.5


class EFGQuaternion(AtomsProperty):

    """
    EFGQuaternion

    Produces a list of ase.Quaternion objects expressing the orientation of
    the EFG tensors with respect to the cartesian axes.
    Requires the Atoms object to have been loaded from a .magres file
    containing the relevant information.

    | Parameters:
    |   force_recalc (bool): if True, always diagonalise the tensors even if
    |                        already present.

    | Returns:
    |   ms_quat (np.ndarray): list of quaternions

    """

    default_name = 'efg_quats'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    @_has_efg_check
    def extract(s, force_recalc):

        if (not s.has(EFGDiagonal.default_name + '_evecs') or
                force_recalc):
            EFGDiagonal.get(s)

        efg_evecs = s.get_array(EFGDiagonal.default_name + '_evecs')

        return _evecs_2_quat(efg_evecs)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of AtomsProperties that relate to linkage of atoms"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
import pkgutil
import numpy as np
from ase.data import atomic_numbers
from ase.quaternions import Quaternion
from soprano.selection import AtomSelection
from soprano.properties import AtomsProperty
from soprano.utils import (swing_twist_decomp, is_string, graph_specsort,
                           minimum_periodic, all_periodic, get_bonding_graph)
from soprano.data import vdw_radii


def _compute_bonds(s, vdw_set, vdw_scale=1.0, default_vdw=2.0, vdw_custom={}):
    """Convenience function that covers the core of bond computation"""

    # First, we need the biggest Van der Waals radius
    # So that we know how big the supercell needs to be

    # Build a custom VdW set
    vdw_r = np.array(vdw_radii[vdw_set])*vdw_scale
    vdw_r = np.where(np.isnan(vdw_r), default_vdw, vdw_r)
    for el, r in vdw_custom.items():
        vdw_r[atomic_numbers[el]] = r

    vdw_vals = vdw_r[s.get_atomic_numbers()]
    vdw_max = max(vdw_vals)

    # Get the interatomic pair distances
    atomn = len(s)
    triui = np.triu_indices(atomn, k=1)
    v = s.get_positions()
    v = (v[:, None, :]-v[None, :, :])[triui]
    # Reduce them
    v, v_i, v_cells = all_periodic(v, s.get_cell(), vdw_max, pbc=s.get_pbc())
    v = np.linalg.norm(v, axis=-1)

    # Now distance and VdW matrices
    vdw_M = ((vdw_vals[None, :]+vdw_vals[:, None])/2.0)[triui]
    link_M = v <= vdw_M[v_i]
    linked = np.where(link_M)

    return linked, triui, v, v_i, v_cells


class LinkageList(AtomsProperty):

    """
    LinkageList

    Produces an array containing the atomic pair distances in a system,
    reduced to their shortest periodic version and sorted min to max.

    | Parameters:
    |   size (int): maximum number of distances to include. If not present,
    |               all of them will be included. If present, arrays will be
    |               cut or padded to reach this size.
    |   return_pairs (bool): if True, return the pairs of atoms to which the
    |                        distances correspond, as a list of tuples of
    |                        indices.

    | Returns:
    |   link_list ([float]): sorted list of interatomic linkage distances
    |   pair_list ([(int, int)]): only if return_pairs is True, list of pairs
    |                             corresponding to the distances

    """

    default_name = 'linkage_list'
    default_params = {
        'size': 0,
        'return_pairs': False
    }

    @staticmethod
    def extract(s, size, return_pairs):
        # Get the interatomic pair distances
        v = s.get_positions()
        v = v[:, None, :]-v[None, :, :]
        pair_inds = np.triu_indices(v.shape[0], k=1)
        v = v[pair_inds]
        # Reduce them
        v, _ = minimum_periodic(v, s.get_cell())
        # And now compile the list
        link_list = np.linalg.norm(v, axis=-1)
        sort_i = np.argsort(link_list)
        link_list = link_list[sort_i]
        if size > 0:
            if link_list.shape[0] >= size:
                link_list = link_list[:size]
            else:
                link_list = np.pad(link_list,
                                   (0, size-link_list.shape[0]),
                                   mode=str('constant'),
                                   constant_values=np.inf)

        if not return_pairs:
            return link_list
        else:
            pairs = list(zip(pair_inds[0][sort_i], pair_inds[1][sort_i]))
            return link_list, pairs


class Bonds(AtomsProperty):

    """
    Bonds

    Produces an array of tuples identifying all bonds existing within the
    system (calculated using Van der Waals radii). The tuples are structured
    as:

    (atom_1, atom_2, atom_2_cell, bond_length)

    with atom_1 and atom_2 being indices and atom_2_cell being an array of
    integers identifying the unit cell to which atom_2 belongs with respect
    to atom_1 (which is assumed to be in (0,0,0), the central cell). This is
    to account for the possibility of course that the bond exists through the
    periodic boundary. WARNING: the possibility of an atom bonding with
    another throughout two different periodic boundaries is not accounted for.

    | Parameters:
    |   vdw_set({ase, jmol, csd}): set of Van der Waals radii to use. Default 
    |                              is csd [S. Alvarez, 2013].
    |   vdw_scale (float): scaling factor to apply to the base Van der Waals
    |                      radii values. Values bigger than one make for more
    |                      tolerant bonds.
    |   default_vdw (float): default Van der Waals radius for species for
    |                        whom no data is available.
    |   vdw_custom (dict): a dictionary of custom Van der Waals radii to use,
    |                      overriding the existing ones, expressed as
    |                      {symbol: radius}.
    |   return_matrix (bool): if True, also return an NxN bonding matrix for
    |                         all N atoms in the system
    |   save_info (bool): if True, save the found bonds (and in case matrix)
    |                     as part of the Atoms object info. By default True.

    | Returns:
    |   bonds([tuple]): list of bonds in the form of 3-tuples structured as
    |                   explained above

    """

    default_name = 'bonds'
    default_params = {
        'vdw_set': 'csd',
        'vdw_scale': 1.0,
        'default_vdw': 2.0,
        'vdw_custom': {},
        'return_matrix': False,
        'save_info': True,
    }

    @staticmethod
    def extract(s, vdw_set, vdw_scale, default_vdw, vdw_custom, return_matrix,
                save_info):

        linked, triui, v, v_i, v_cells = _compute_bonds(s,
                                                        vdw_set,
                                                        vdw_scale,
                                                        default_vdw,
                                                        vdw_custom)

        bonds = list(zip(triui[0][v_i[linked]], triui[1][v_i[linked]],
                         -v_cells[linked], v[linked]))

        if save_info:
            s.info[Bonds.default_name] = list(bonds)

        if not return_matrix:
            return list(bonds)  # For Python 3 compatibility
        else:
            bmat = np.zeros((len(s), len(s))).astype(int)
            bmat[triui[0][v_i[linked]], triui[1][v_i[linked]]] = 1
            bmat[triui[1][v_i[linked]], triui[0][v_i[linked]]] = 1

            if save_info:
                s.info[Bonds.default_name + '_matrix'] = bmat

            return list(bonds), bmat


class CoordinationHistogram(AtomsProperty):

    """
    CoordinationHistogram

    Produces an histogram representing, for each pair of species present in
    the system, how many atoms of species 1 have n bonds with species 2, n
    being the histogram bins. The histogram is topped at a 'maximum
    coordination' parameter which is 6 by default but can be user defined;
    the last bin represents all higher values (so by default '6 or more').
    Two species or lists of species can be given if one wants to restrict the
    search; otherwise a full histogram for all pairs of species is returned.

    | Parameters:
    |   vdw_set({ase, jmol, csd}): set of Van der Waals radii to use. Default 
    |                              is csd [S. Alvarez, 2013].
    |   vdw_scale (float): scaling factor to apply to the base Van der Waals
    |                      radii values. Values bigger than one make for more
    |                      tolerant bonds.
    |   default_vdw (float): default Van der Waals radius for species for
    |                        whom no data is available.
    |   vdw_custom (dict): a dictionary of custom Van der Waals radii to use,
    |                      overriding the existing ones, expressed as
    |                      {symbol: radius}.
    |   species_1 (str or [str]): list of species to compute the histogram
    |                             for. By default all of them.
    |   species_2 (str or [str]): list of species whose coordination with
    |                             species_1 should be checked. By default all
    |                             of them.
    |   max_coord (int): what should be the largest coordination number
    |                    considered for an atom (default 6).

    | Returns:
    |   coord_hist (dict): dictionary of dictionaries indexed by species_1
    |                      followed by species_2. The elements are arrays of
    |                      integers constituting the histogram.

    """

    default_name = 'coord_histogram'
    default_params = {
        'vdw_set': 'csd',
        'vdw_scale': 1.0,
        'default_vdw': 2.0,
        'vdw_custom': {},
        'species_1': None,
        'species_2': None,
        'max_coord': 6
    }

    @staticmethod
    def extract(s, vdw_set, vdw_scale, default_vdw, vdw_custom,
                species_1, species_2, max_coord):

        elems = np.array(s.get_chemical_symbols())

        # Get the bonds
        bond_calc = Bonds(vdw_set=vdw_set,
                          vdw_scale=vdw_scale,
                          default_vdw=default_vdw,
                          vdw_custom=vdw_custom)
        bonds = bond_calc(s)
        # What if there are none?
        if len(bonds) == 0:
            # Just return
            print('WARNING: no bonds detected for CoordinationHistogram')
            return {}
        bond_inds = np.concatenate(list(zip(*bonds))[:2])
        bond_elems = elems[bond_inds]
        bN = len(bonds)

        if species_1 is None:
            species_1 = np.unique(elems)
        elif is_string(species_1):
            species_1 = np.array([species_1])

        if species_2 is None:
            species_2 = np.unique(elems)
        elif is_string(species_2):
            species_2 = np.array([species_2])

        # Initialise the histogram
        hist = {s1: {s2: np.zeros(max_coord+1)
                     for s2 in species_2}
                for s1 in species_1}

        for s1 in species_1:
            # Which atoms are of species 1, and what are they bonded to?
            i1 = np.where(bond_elems == s1)[0]
            b1 = bond_inds[i1]
            be1 = bond_elems[(i1-bN).astype(int)]
            for s2 in species_2:
                # Which ones are bonded to species 2?
                i2 = np.where(be1 == s2)
                b2 = b1[i2]
                b2, counts = np.unique(b2, return_counts=True)
                hist_i, hist_n = np.unique(counts, return_counts=True)
                # Fix for numbers that are too high...
                hist_big = np.where(hist_i > max_coord)[0]
                if (len(hist_big) > 0):
                    # In this case find the max_coord index, if absent add it
                    hist_maxc = np.where(hist_i == max_coord)[0]
                    if len(hist_maxc) == 0:
                        hist_i = np.concatenate([hist_i, [max_coord]])
                        hist_n = np.concatenate([hist_n, [0]])
                        hist_maxc = [-1]
                    hist_n[hist_maxc] += np.sum(hist_n[hist_big])
                    # Then slice away, keep only the admissible indices
                    hist_small = np.where(hist_i <= max_coord)[0]
                    hist_i = hist_i[hist_small]
                    hist_n = hist_n[hist_small]
                hist[s1][s2][hist_i] += hist_n

        return hist


class Molecules(AtomsProperty):

    """
    Molecules

    Produces an array containing multiple AtomSelection objects representing
    molecules in the system as found by connecting atoms closer than the half
    sum of their Van der Waals radii. It will return the entire unit cell if
    the system can not be split in molecules at all.

    | Parameters:
    |   vdw_set({ase, jmol, csd}): set of Van der Waals radii to use. Default 
    |                              is csd [S. Alvarez, 2013].
    |   vdw_scale (float): scaling factor to apply to the base Van der Waals
    |                      radii values. Values bigger than one make for more
    |                      tolerant molecules.
    |   default_vdw (float): default Van der Waals radius for species for
    |                        whom no data is available.
    |   vdw_custom (dict): a dictionary of custom Van der Waals radii to use,
    |                      overriding the existing ones, expressed as
    |                      {symbol: radius}.
    |   save_info (bool): if True, save the found molecules as part of the
    |                     Atoms object info. By default True.

    | Returns:
    |   molecules ([AtomSelection]): list of molecules in the form of
    |                                AtomSelection objects.

    """

    default_name = 'molecules'
    default_params = {
        'vdw_set': 'csd',
        'vdw_scale': 1.0,
        'default_vdw': 2.0,
        'vdw_custom': {},
        'save_info': True,
    }

    @staticmethod
    def extract(s, vdw_set, vdw_scale, default_vdw, vdw_custom, save_info):

        N = len(s)
        # Sanity check
        if N < 2:
            # WTF?
            print('WARNING: impossible to calculate molecules on single-atom '
                  'system')
            return None

        # Get the bonds
        bond_calc = Bonds(vdw_set=vdw_set,
                          vdw_scale=vdw_scale,
                          default_vdw=default_vdw,
                          vdw_custom=vdw_custom)
        bonds = bond_calc(s)

        mol_sets = []
        unsorted_atoms = list(range(N))

        def get_linked(i):
            i_bonds = filter(lambda b: i in b[:2], bonds)
            links = map(lambda b: (b[1], b[2]) if b[0] == i else (b[0], -b[2]),
                        i_bonds)
            return links

        while len(unsorted_atoms) > 0:
            mol_queue = [(unsorted_atoms.pop(0), np.zeros(3))]
            current_mol = []
            current_mol_cells = []
            current_mol_bonds = []
            while len(mol_queue) > 0:
                a1, cell1 = mol_queue.pop(0)
                current_mol.append(a1)
                current_mol_cells.append(cell1)
                current_mol_bonds.append([])
                # Find linked atoms
                links = get_linked(a1)
                for l, cl in links:
                    if l in unsorted_atoms:
                        mol_queue.append((l, cell1 + cl))
                        unsorted_atoms.remove(l)
                    current_mol_bonds[-1].append(l)

            mol_sets.append((current_mol, current_mol_cells,
                             current_mol_bonds))

        mols = []
        for m_i, m_cells, m_bonds in mol_sets:
            mols.append(AtomSelection(s, m_i))
            mols[-1].set_array('cell_indices', m_cells)
            # This is necessary to guarantee shape consistency
            m_barr = np.empty((len(m_bonds),), dtype=list)
            for i, m_b in enumerate(m_bonds):
                m_barr[i] = m_b
            mols[-1].set_array('bonds', m_barr)

        if save_info:
            s.info[Molecules.default_name] = mols

        return mols


class MoleculeNumber(AtomsProperty):

    """
    MoleculeNumber

    Number of molecules detected in this system. By default will use already
    existing molecules if they're present as a saved array in the system.


    | Parameters:
    |   force_recalc (bool): if True, always recalculate the molecules even if
    |                        already present.

    | Returns:
    |   molecule_n (int): number of molecules found

    """

    default_name = 'molecule_n'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    def extract(s, force_recalc):

        if Molecules.default_name not in s.info or force_recalc:
            Molecules.get(s)

        return len(s.info[Molecules.default_name])


class MoleculeMass(AtomsProperty):

    """
    MoleculeMass

    Total mass of each of the molecules detected in this system. By default
    will use already existing molecules if they're present as a saved array in
    the system.


    | Parameters:
    |   force_recalc (bool): if True, always recalculate the molecules even if
    |                        already present.
    |   size (int): maximum number of distances to include. If not present,
    |               all of them will be included. If present, arrays will be
    |               cut or padded to reach this sizeber.

    | Returns:
    |   molecule_m ([float]): mass of each of the molecules present, sorted.

    """

    default_name = 'molecule_mass'
    default_params = {
        'force_recalc': False,
        'size': 0,
    }

    @staticmethod
    def extract(s, force_recalc, size):

        if Molecules.default_name not in s.info or force_recalc:
            Molecules.get(s)

        mol_m = []
        all_m = s.get_masses()
        for mol in s.info[Molecules.default_name]:
            mol_m.append(np.sum(all_m[mol.indices]))

        mol_m = np.array(mol_m)
        mol_m.sort()

        if size > 0:
            if mol_m.shape[0] >= size:
                mol_m = mol_m[:size]
            else:
                mol_m = np.pad(mol_m,
                               (0, size-mol_m.shape[0]),
                               mode=str('constant'),
                               constant_values=np.inf)

        return mol_m


class MoleculeCOM(AtomsProperty):

    """
    MoleculeCOM

    List of centers of mass for the molecules present in the system. By
    default will use already existing molecules if they're present as a saved
    array in the system.

    | Parameters:
    |   force_recalc (bool): if True, always recalculate the molecules even if
    |                        already present.

    | Return:
    |   mol_com (np.ndarray): list of centers of mass for the system's
    |                         molecules

    """

    default_name = 'molecule_com'
    default_params = {
        'force_recalc': False,
    }

    @staticmethod
    def extract(s, force_recalc):

        if Molecules.default_name not in s.info or force_recalc:
            Molecules.get(s)

        mol_com = []
        all_m = s.get_masses()
        all_pos = s.get_positions()

        for mol in s.info[Molecules.default_name]:

            mol_pos = all_pos[mol.indices]
            mol_pos += np.tensordot(mol.get_array('cell_indices'),
                                    s.get_cell(),
                                    axes=(1, 0))
            mol_ms = all_m[mol.indices]
            mol_com.append(np.sum(mol_pos*mol_ms[:, None],
                                  axis=0)/np.sum(mol_ms))

        return np.array(mol_com)


class MoleculeCOMLinkage(AtomsProperty):

    """
    MoleculeCOMLinkage

    Linkage list - following the same criteria as the atomic one - calculated
    for the centers of mass of the molecules present in the system. By default
    will use already existing molecules if they're present as a saved array in
    the system.


    | Parameters:
    |   force_recalc (bool): if True, always recalculate the molecules even if
    |                        already present.
    |   size (int): maximum number of distances to include. If not present,
    |               all of them will be included. If present, arrays will be
    |               cut or padded to reach this sizeber.

    | Returns:
    |   molecule_linkage ([float]): distances between all centers of mass of
    |                               molecules in the system, sorted.

    """

    default_name = 'molecule_com_linkage'
    default_params = {
        'force_recalc': False,
        'size': 0,
    }

    @staticmethod
    def extract(s, force_recalc, size):

        if Molecules.default_name not in s.info or force_recalc:
            Molecules.get(s)

        mol_com = MoleculeCOM.get(s)

        # Safety check
        if len(mol_com) < 2:
            return [np.inf]*size

        # Now make the linkage
        v = np.array(mol_com)
        v = v[:, None, :]-v[None, :, :]
        v = v[np.triu_indices(v.shape[0], k=1)]
        # Reduce them
        v, _ = minimum_periodic(v, s.get_cell())
        # And now compile the list
        link_list = np.linalg.norm(v, axis=-1)
        link_list.sort()

        if size > 0:
            if link_list.shape[0] >= size:
                link_list = link_list[:size]
            else:
                link_list = np.pad(link_list,
                                   (0, size-link_list.shape[0]),
                                   mode=str('constant'),
                                   constant_values=np.inf)

        return link_list


class MoleculeQuaternion(AtomsProperty):

    """
    MoleculeQuaternion

    A list of quaternions expressing the rotation of the molecule's intertia
    tensor principal frame with respect to the cartesian axes.

    | Parameters:
    |   force_recalc (bool): if True, always recalculate the molecules even if
    |                        already present.

    | Returns:
    |   mol_quat ([ase.Quaternion]): list of quaternions

    """

    default_name = 'molecule_quaternion'
    default_params = {
        'force_recalc': False,
    }

    @staticmethod
    def extract(s, force_recalc):

        if Molecules.default_name not in s.info or force_recalc:
            Molecules.get(s)

        mol_quat = []
        all_m = s.get_masses()
        all_pos = s.get_positions()

        for mol in s.info[Molecules.default_name]:

            mol_pos = all_pos[mol.indices]
            mol_pos += np.tensordot(mol.get_array('cell_indices'),
                                    s.get_cell(),
                                    axes=(1, 1))
            mol_ms = all_m[mol.indices]

            # We still need to correct the positions with the COM
            mol_com = np.sum(mol_pos*mol_ms[:, None],
                             axis=0)/np.sum(mol_ms)
            mol_pos -= mol_com

            tens_i = np.identity(3)[None, :, :] * \
                np.linalg.norm(mol_pos, axis=1)[:, None, None]**2

            tens_i -= mol_pos[:, None, :]*mol_pos[:, :, None]
            tens_i *= mol_ms[:, None, None]
            tens_i = np.sum(tens_i, axis=0)

            evals, evecs = np.linalg.eigh(tens_i)
            # General ordering convention: we want the component of the
            # longest position to be positive along evecs_0, and the component
            # of the second longest (and non-parallel) position to be positive
            # along evecs_1, and the triple to be right-handed of course.
            mol_pos = sorted(mol_pos, key=lambda x: -np.linalg.norm(x))
            if len(mol_pos) > 1:
                evecs[0] *= np.sign(np.dot(evecs[0], mol_pos[0]))
            e1dirs = np.where(np.linalg.norm(np.cross(mol_pos,
                                                      mol_pos[0])) > 0)[0]
            if len(e1dirs) > 0:
                evecs[1] *= np.sign(np.dot(evecs[1], mol_pos[e1dirs[0]]))
            evecs[2] *= np.sign(np.dot(evecs[2],
                                       np.cross(evecs[0], evecs[1])))
            # Evecs must be proper
            evecs /= np.linalg.det(evecs)

            quat = Quaternion()
            quat = quat.from_matrix(evecs.T)
            mol_quat.append(quat)

        return mol_quat


class MoleculeRelativeRotation(AtomsProperty):

    """
    MoleculeRelativeRotation

    A list of relative rotations between molecules. Uses the inertia tensor
    eigenvectors to establish a local frame for each molecule, then uses
    quaternions to define a rotational distance between molecules. It then
    produces a list of geodesic distances between these quaternions.


    | Parameters:
    |   force_recalc (bool): if True, always recalculate the molecules even if
    |                        already present.
    |   size (int): maximum number of distances to include. If not present,
    |               all of them will be included. If present, arrays will be
    |               cut or padded to reach this size.
    |   twist_axis ([float]): if present, only compare the Twist component of
    |                         quaternion along the given axis. The Twist/Swing
    |                         decomposition splits a quaternion in a rotation
    |                         around an axis and one around an orthogonal
    |                         direction. Only one between this and swing_plane
    |                         can be present.
    |   swing_plane ([float]): if present, only compare the Swing component of
    |                         quaternion along the given axis. The Twist/Swing
    |                         decomposition splits a quaternion in a rotation
    |                         around an axis and one around an orthogonal
    |                         direction. Only one between this and twist_axis
    |                         can be present.

    | Returns:
    |   molecule_relrot ([float]): list of relative rotations, as quaternion
    |                              distances, with the required ordering.

    """

    default_name = 'molecule_rel_rotation'
    default_params = {
        'force_recalc': False,
        'size': 0,
        'swing_plane': None,
        'twist_axis': None,
    }

    @staticmethod
    def extract(s, force_recalc, size, swing_plane, twist_axis):

        # Sanity check
        if swing_plane is not None and twist_axis is not None:
            raise RuntimeError('Only one between swing_plane and twist_axis '
                               'can be passed to MoleculeRelativeRotation')

        if Molecules.default_name not in s.info or force_recalc:
            Molecules.get(s)

        mol_quat = [q.q for q in MoleculeQuaternion.get(s)]

        for i, quat in enumerate(mol_quat):
            if swing_plane is not None:
                mol_quat[i], dummy = swing_twist_decomp(quat, swing_plane)
            elif twist_axis is not None:
                dummy, mol_quat[i] = swing_twist_decomp(quat, twist_axis)

        # Safety check
        if len(mol_quat) < 2:
            return [np.inf]*size

        # Now make the linkage
        v = np.array(mol_quat)
        v = np.tensordot(v, v, axes=(-1, -1))
        link_list = 1.0-np.abs(v[np.triu_indices(v.shape[0], k=1)])
        link_list.sort()

        if size > 0:
            if link_list.shape[0] >= size:
                link_list = link_list[:size]
            else:
                link_list = np.pad(link_list,
                                   (0, size-link_list.shape[0]),
                                   mode=str('constant'),
                                   constant_values=np.inf)

        return np.array(link_list)


class MoleculeSpectralSort(AtomsProperty):

    """
    MoleculeSpectralSort

    Reorder molecules to have their indices sorted using a spectral 
    sorting method based on the Fiedler vector of their bonding graph. This
    sorting should be equivalent for equivalent molecules - except for the
    arbitrary ordering of equivalent atoms.

    | Parameters:
    |   force_recalc (bool): if True, always recalculate the molecules even if
    |                        already present.

    | Returns:
    |   mol_specsort ([np.ndarray]): list of Molecules with indices sorted by
                                     spectral method.

    """

    default_name = 'molecule_specsort'
    default_params = {
        'force_recalc': False,
    }

    @staticmethod
    def extract(s, force_recalc):

        if Molecules.default_name not in s.info or force_recalc:
            Molecules.get(s)

        mol_specsort = []

        for mol in s.info[Molecules.default_name]:
            # Start by getting the adjacency and degree matrices
            N = len(mol)
            bonds = mol.get_array('bonds')
            A = np.zeros((N, N))
            D = A.copy()
            for i, b in enumerate(bonds):
                D[i, i] = len(b)
                A[i, [list(mol.indices).index(j) for j in b]] = 1

            L = D - A

            sort = graph_specsort(L)

            mol = mol[sort]
            mol_specsort.append(mol)

        return mol_specsort


class HydrogenBonds(AtomsProperty):

    """
    Hydrogen Bonds

    Produces a dictionary containing the atom indices defining hydrogen bonds
    detected in the system - if required, classified by type. By default only
    O and N atoms are considered for hydrogen bonds (OH..O, OH..N and so on).
    The type is defined as AH..B where A is the symbol of the atom directly
    bonded to the proton and B the one of the hydrogen bonded one.

    | Parameters:
    |   vdw_set({ase, jmol, csd}): set of Van der Waals radii to use. Default 
    |                              is csd [S. Alvarez, 2013].
    |   vdw_scale (float): scaling factor to apply to the base Van der Waals
    |                      radii values. Values bigger than one make for more
    |                      tolerant molecules.
    |   default_vdw (float): default Van der Waals radius for species for
    |                        whom no data is available.
    |   hbond_elems ([str]): chemical symbols of elements considered capable
    |                        of forming hydrogen bonds (by default O and N)
    |   max_length (float): maximum A-B length of the hydrogen bond in
    |                       Angstrom - default is 3.5 Ang
    |   max_angle (float): maximum A-H/A-B angle in the hydrogen bond in
    |                      degrees - default is 45 deg
    |   save_info (bool): if True, save the found hydrogen bonds as part of
    |                     the Atoms object info. By default True.


    | Returns:
    |   hbondss ([dict]): list of hydrogen bonds detected
    |                     in the system by type (can contain empty arrays).
    |                     For each hydrogen bond we give index of the H
    |                     atom, index and unit cell of the A atom (the one
    |                     directly bonded), index and unit cell of the B atom
    |                     (the one that's hydrogen bonded), length and angle
    |                     in degrees.

    """

    default_name = 'hydrogen_bonds'
    default_params = {
        'vdw_set': 'csd',
        'vdw_scale': 1.0,
        'default_vdw': 2.0,
        'vdw_custom': {},
        'hbond_elems': ['O', 'N'],
        'max_length': 3.5,
        'max_angle': 45.0,
        'save_info': True
    }

    @staticmethod
    def extract(s, vdw_set, vdw_scale, default_vdw, vdw_custom, hbond_elems,
                max_length, max_angle, save_info):

        def elem_inds(s, el):
            return [i for i, cs in enumerate(s.get_chemical_symbols())
                    if cs == el]

        def bname(A, B):
            return '{0}H..{1}'.format(A, B)

        # Define types
        hbonds = {}
        for elA in hbond_elems:
            for elB in hbond_elems:
                hbonds[bname(elA, elB)] = []

        # First, grab the hydrogen atoms
        h_atoms = elem_inds(s, 'H')
        if len(h_atoms) == 0:
            # Nothing to do
            if save_info:
                s.info[HydrogenBonds.default_name] = hbonds
            return hbonds

        bond_atoms = []
        for el_b in hbond_elems:
            bond_atoms += elem_inds(s, el_b)

        if len(bond_atoms) < 2:
            if save_info:
                s.info[HydrogenBonds.default_name] = hbonds
            return hbonds

        # Now to pick the positions
        h_atoms_pos = s.get_positions()[h_atoms]
        bond_atoms_pos = s.get_positions()[bond_atoms]
        # Van der Waals radii length of H-atom bonds

        vdw_r = np.array(vdw_radii[vdw_set])*vdw_scale
        vdw_r = np.where(np.isnan(vdw_r), default_vdw, vdw_r)
        for el, r in vdw_custom.items():
            vdw_r[atomic_numbers[el]] = r

        bonds_vdw = vdw_r[s.get_atomic_numbers()[bond_atoms]]
        bonds_vdw = (bonds_vdw+vdw_r[1])/2.0

        # Now find the shortest and second shortest bonds for each H
        h_links = (h_atoms_pos[:, None, :]-bond_atoms_pos[None, :, :])
        shape = h_links.shape
        h_links, h_cells = minimum_periodic(h_links.reshape((-1, 3)),
                                            s.get_cell())
        h_links = h_links.reshape(shape)
        h_cells = h_cells.reshape(shape)
        h_links_norm = np.linalg.norm(h_links, axis=-1)

        # Now for each hydrogen: first and second closest
        h_closest = np.argsort(h_links_norm,
                               axis=-1)[:, :2]

        # Which ones DO actually form bonds?
        rngh = range(len(h_atoms))
        # Condition one: closest atom, A, is bonded
        h_bonded = h_links_norm[rngh,
                                h_closest[:, 0]] <= bonds_vdw[h_closest[:, 0]]
        # Condition two: furthest atom, B, is NOT bonded...
        h_bonded = np.logical_and(h_bonded,
                                  h_links_norm[rngh,
                                               h_closest[:, 1]
                                               ] > bonds_vdw[h_closest[:, 1]])
        # Condition three: ...but still closer to A than max_length
        links_ab = h_links[rngh, h_closest[:, 0]] - \
            h_links[rngh, h_closest[:, 1]]
        links_ab_norm = np.linalg.norm(links_ab, axis=-1)
        h_bonded = np.logical_and(h_bonded, links_ab_norm <= max_length)
        # Condition four: finally, the angle between AH and AB in A-H..B
        # must be smaller than max_angle
        angles_abah = np.sum(links_ab*h_links[rngh, h_closest[:, 0]], axis=-1)
        angles_abah /= links_ab_norm*h_links_norm[rngh, h_closest[:, 0]]
        angles_abah = np.arccos(angles_abah)*180.0/np.pi
        h_bonded = np.logical_and(h_bonded, angles_abah <= max_angle)

        # The survivors are actual h bonds!
        # Now on to defining them
        h_bonded = np.where(h_bonded)[0]
        if len(h_bonded) == 0:
            if save_info:
                s.info[HydrogenBonds.default_name] = hbonds
            return hbonds

        # Moving to structure indices

        for h in h_bonded:

            ai, bi = h_closest[h]
            elA = s.get_chemical_symbols()[bond_atoms[ai]]
            elB = s.get_chemical_symbols()[bond_atoms[bi]]

            bond = {'H': h_atoms[h],
                    'A': (bond_atoms[ai], h_cells[h, ai]),
                    'B': (bond_atoms[bi], h_cells[h, bi]),
                    'length': links_ab_norm[h],
                    'angle': angles_abah[h]}

            btype = bname(elA, elB)
            hbonds[btype].append(bond)

        if save_info:
            s.info[HydrogenBonds.default_name] = hbonds

        return hbonds


class HydrogenBondsNumber(AtomsProperty):

    """
    HydrogenBondsNumber

    Number of hydrogen bonds detected in this system, classified by type.
    By default will use already existing hydrogen bonds if they're present as
    a saved array in the system.


    | Parameters:
    |   force_recalc (bool): if True, always recalculate the hydrogen bonds
    |                        even if already present.

    | Returns:
    |   hbonds_n (int): number of hydrogen bonds found

    """

    default_name = 'hydrogen_bonds_n'
    default_params = {
        'force_recalc': False
    }

    @staticmethod
    def extract(s, force_recalc):

        if HydrogenBonds.default_name not in s.info or force_recalc:
            hbonds = HydrogenBonds.get(s)
        else:
            hbonds = s.info[HydrogenBonds.default_name]

        hbonds_n = {}
        for hbt in hbonds:
            hbonds_n[hbt] = len(hbonds[hbt])

        return hbonds_n


class DihedralAngleList(AtomsProperty):

    """
    DihedralAngleList

    Produces a list of dihedral angles found in the system, identified by
    looking for a bonding pattern. The amount of said angles can vary from
    zero (if the pattern is not present) to an arbitrary number. They will be
    returned sorted from lowest to highest. Periodic boundary conditions are
    taken into account. If no pattern is provided, HCCH groups are searched by
    default.

    | Parameters:
    |   dihedral_pattern ([str]*4): a list of four chemical symbols
    |                               identifying the dihedral pattern to look
    |                               for
    |   bonds_params (dict): parameters to pass to the Bonds property used to
    |                        compute bonds. See the Bonds docstring for
    |                        details. If not provided, defaults are used

    | Returns:
    |   dihedral_angles (np.ndarray): sorted list of dihedral angles found

    """

    default_name = 'dihedral_angle_list'
    default_params = {
        'dihedral_pattern': ['H', 'C', 'C', 'H'],
        'bonds_params': {}
    }

    @staticmethod
    def extract(s, dihedral_pattern, bonds_params):

        # First, compute bonds
        bonds = Bonds(**bonds_params)
        bonds = bonds(s)
        # Chemical symbols
        elems = np.array(s.get_chemical_symbols())

        # Build a network
        dp = np.array(dihedral_pattern)
        nodes = np.where(np.any(elems[:, None] == dp[None, :],
                                axis=1))[0]

        bond_table = {n: [] for n in nodes}
        for b in bonds:
            if np.all(np.any(nodes[:, None] == b[:2], axis=0)):
                bond_table[b[0]].append((b[1], b[2], b[3]))
                bond_table[b[1]].append((b[0], -b[2], b[3]))

        # Now prepare a series of 'pointers' to traverse the network
        pointers = np.where(elems == dihedral_pattern[0])[0]
        pointer_memory = [[] for p in pointers]
        pointer_traversal = np.zeros((len(pointers), 1, 3))
        for el in dihedral_pattern[1:]:
            new_pointers = []
            new_pointer_memory = []
            new_pointer_traversal = []
            for p_i, p in enumerate(pointers):
                # Is it bonded to anything with that element?
                # Did we visit it already?
                for b in bond_table[p]:
                    if elems[b[0]] == el and b[0] not in pointer_memory[p_i]:
                        # If so, spawn new pointers and save the corresponding
                        # memory of the previous path
                        new_pointers.append(b[0])
                        new_pointer_memory.append(
                            list(pointer_memory[p_i]) + [p])
                        new_pointer_traversal.append(
                            np.concatenate((pointer_traversal[p_i],
                                            np.array(b[1])[None, :])))
            pointers = new_pointers
            pointer_memory = new_pointer_memory
            pointer_traversal = np.array(new_pointer_traversal)

        # Now build the dihedra array
        dihedra = np.array([pointer_memory[i] + [p]
                            for i, p in enumerate(pointers)])

        # Symmetry check, epurate the repeated ones if needed
        is_symm = np.all(dihedral_pattern == dihedral_pattern[::-1])

        if is_symm:
            duplicate = []
            for i, d1 in enumerate(dihedra):
                for j, d2 in enumerate(dihedra[i+1:]):
                    if np.all(d2 == d1[::-1]):
                        duplicate.append(j+i+1)
            unique = list(set(range(len(dihedra))) - set(duplicate))
            dihedra = dihedra[unique]
            pointer_traversal = pointer_traversal[unique]

        # If it's empty we may quit now
        if len(dihedra) == 0:
            return np.array([])
        # Cumulate traversals
        pointer_traversal = np.cumsum(pointer_traversal, axis=1)

        # Now that we have them let's find the positions
        posv = s.get_positions()[dihedra.reshape((-1,))]
        posv += np.dot(pointer_traversal.reshape((-1, 3)), s.get_cell())

        # And finally the dihedral angles
        links = np.diff(posv.reshape((-1, 4, 3)), axis=1)
        bxa = np.cross(links[:, 1, :], links[:, 0, :])
        bxa /= np.linalg.norm(bxa, axis=-1)[:, None]
        cxb = np.cross(links[:, 2, :], links[:, 1, :])
        cxb /= np.linalg.norm(cxb, axis=-1)[:, None]
        angles = np.arccos(np.clip(np.sum(bxa*cxb, axis=-1), -1, 1))
        angles = np.where(np.sum(bxa*links[:, 2, :], axis=-1) > 0,
                          2*np.pi-angles, angles)

        # And return!
        return angles


class BondGraph(AtomsProperty):
    """
    BondGraph

    Bond graph returns a networkx graph of the molecules in the structure. To
    use this property you must have the networkx library installed.

    | Parameters:
    |   force_recalc (bool): if True, always recalculate the bond graph
    |                        even if already present.
    |   save_info (bool): if True, save the bond graph as part of the Atoms
    |                     object info. By default True.

    | Returns:
    |   graph (nx.Graph): the bond graph for the structure

    """

    default_name = 'bond_graph'
    default_params = {
        'force_recalc': False,
        'save_info': True,
    }

    @staticmethod
    def extract(s, force_recalc, save_info):
        if BondGraph.default_name not in s.info or force_recalc:
            bprop = Bonds(return_matrix=True)
            _, bond_matrix = bprop(s)
            graph = get_bonding_graph(bond_matrix)
        else:
            graph = s.info[BondGraph.default_name]

        if save_info:
            s.info[BondGraph.default_name] = graph

        return graph



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes and functions for simulating approximated NMR spectroscopic results
from structures.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import json
import pkgutil
import numpy as np
from ase import Atoms
from scipy.special import fresnel
from scipy import constants as cnst
from collections import namedtuple
from soprano.utils import minimum_supcell, supcell_gridgen
from soprano.nmr.utils import _dip_constant
from soprano.data.nmr import (_get_isotope_data, _get_nmr_data, _el_iso,
                              EFG_TO_CHI)
from soprano.calculate.powder.triavg import TriAvg
from soprano.properties.nmr import DipolarCoupling
from soprano.selection import AtomSelection

_nmr_data = _get_nmr_data()

# Conversion functions to Tesla
# (they take element and isotope as arguments)
_larm_units = {
    'MHz': lambda e, i: 2*np.pi*1.0e6/_nmr_data[e][i]['gamma'],
    'T': lambda e, i: 1.0,
}

# Function used for second-order quadrupolar shift
# Arguments: cos(2*alpha), eta


def _st_A(c2a, eta):
    return -27.0/8.0-9.0/4.0*eta*c2a-3.0/8.0*eta**2*c2a**2


def _st_B(c2a, eta):
    return 15.0/4.0-0.5*eta**2+2*eta*c2a+0.75*eta**2*c2a**2


def _st_C(c2a, eta):
    return -23.0/40.0+14.0/15.0*eta**2+0.25*eta*c2a-3.0/8.0*eta**2*c2a**2


def _mas_A(c2a, eta):
    return 21.0/16.0-7.0/8.0*eta*c2a+7.0/48.0*eta**2*c2a**2


def _mas_B(c2a, eta):
    return -9.0/8.0+1.0/12.0*eta**2+eta*c2a-7.0/24.0*eta**2*c2a**2


def _mas_C(c2a, eta):
    return 9.0/80.0-1.0/15.0*eta**2-0.125*eta*c2a+7.0/48.0*eta**2*c2a**2


def _gfunc(ca, cb, eta, A, B, C):
    c2a = 2*ca**2-1
    return A(c2a, eta)*cb**4+B(c2a, eta)*cb**2+C(c2a, eta)


# Flags for what to include in spectra
NMRFlags = namedtuple('NMRFlags',
                      """CS_ISO
                      CS_ORIENT
                      CS
                      Q_1_ORIENT
                      Q_2_SHIFT
                      Q_2_ORIENT_STATIC
                      Q_2_ORIENT_MAS
                      Q_2_STATIC
                      Q_2_MAS
                      Q_STATIC
                      Q_MAS
                      STATIC
                      MAS"""
                      )
NMRFlags = NMRFlags(
    CS_ISO=1,
    CS_ORIENT=2,
    CS=1+2,
    Q_1_ORIENT=4,
    Q_2_SHIFT=8,
    Q_2_ORIENT_STATIC=16,
    Q_2_ORIENT_MAS=32,
    Q_2_STATIC=8+16,
    Q_2_MAS=8+32,
    Q_STATIC=4+8+16,
    Q_MAS=8+32,
    STATIC=1+2+4+8+16,
    MAS=1+8+32)


class NMRCalculator(object):

    """NMRCalculator

    An object providing an interface to produce basic simulated NMR spectra
    from .magres files. It should be kept in mind that this is *not* a proper
    spin simulation tool, but merely provides a 'guide for the eye' kind of
    spectrum to compare to experimental results. What it can simulate:

    - chemical shift of NMR peaks
    - quadrupolar shifts of NMR peaks up to second order corrections
    - effects of crystal orientation (single crystal)
    - powder average (policrystalline/powder)
    - ultrafast MAS limit spectra

    What it can NOT simulate:

    - finite speed MAS spectra
    - J couplings
    - dipolar interactions
    - complex NMR experiments

    A list of the currently available NMRFlags to be used in conjunction with
    methods that require a list of effects of interest: ::

        NMRFlags.CS_ISO     => chemical shielding, isotropic effect
                .CS_ORIENT  => chemical shielding, orientation dependent
                               effects
                .CS         => chemical shielding, everything
                .Q_1_ORIENT => quadrupolar, 1st order, orientation dependent
                               effects
                .Q_2_SHIFT  => quadrupolar, 2nd order, isotropic shift
                .Q_2_ORIENT_STATIC => quadrupolar, 2nd order, orientation
                                      dependent effects; static limit
                .Q_2_ORIENT_MAS => quadrupolar, 2nd order, orientation
                                      dependent effects; ultrafast MAS limit
                .Q_2_STATIC => quadrupolar, 2nd order, all static effects
                .Q_2_MAS    => quadrupolar, 2nd order, all ultrafast MAS
                               effects
                .Q_STATIC   => quadrupolar, all static effects
                .Q_MAS      => quadrupolar, all ultrafast MAS effects
                .STATIC     => all static effects
                .MAS        => all ultrafast MAS effects


    | Args:
    |   sample (ase.Atoms): an Atoms object describing the system to simulate
    |                       on. Should be loaded with ASE from a .magres file
    |                       if data on shieldings and EFGs is necessary. It
    |                       can also have an optional 'isotopes' array. If it
    |                       does, it will be used in the set_isotopes method
    |                       and interpreted as described in its documentation.
    |   larmor_frequency (float): larmor frequency of the virtual spectrometer
    |                             (referenced to Hydrogen). Default is 400.
    |   larmor_units (str): units in which the larmor frequency is expressed.
    |                       Default are MHz.

    """

    def __init__(self, sample, larmor_frequency=400,
                 larmor_units='MHz'):

        if not isinstance(sample, Atoms):
            raise TypeError('sample must be an ase.Atoms object')

        self._sample = sample

        # Define isotope array
        self._elems = np.array(self._sample.get_chemical_symbols())
        if self._sample.has('isotopes'):
            isos = self._sample.get_array('isotopes')
        else:
            isos = [None]*len(self._sample)
        self.set_isotopes(isos)

        self.set_larmor_frequency(larmor_frequency, larmor_units)

        self._references = {}

    def set_larmor_frequency(self, larmor_frequency=400, larmor_units='MHz',
                             element='1H'):
        """
        Set the Larmor frequency of the virtual spectrometer with the desired
        units and reference element.

        | Args:
        |   larmor_frequency (float): larmor frequency of the virtual
        |                             spectrometer. Default is 400.
        |   larmor_units (str): units in which the larmor frequency is
        |                       expressed. Can be MHz or T. Default are MHz.
        |   element (str): element and isotope to reference the frequency to.
        |                  Should be in the form <isotope><element>. Isotope
        |                  is optional, if absent the most abundant NMR active
        |                  one will be used. Default is 1H.

        """

        if larmor_units not in _larm_units:
            raise ValueError('Invalid units for Larmor frequency')

        # Split isotope and element
        el, iso = _el_iso(element)

        self._B = larmor_frequency*_larm_units[larmor_units](el, iso)

    def get_larmor_frequency(self, element):
        """
        Get the Larmor frequency of the virtual spectrometer for the desired
        element in MHz.

        | Args:
        |   element (str): element and isotope whose frequency we require.
        |                  Should be in the form <isotope><element>. Isotope
        |                  is optional, if absent the most abundant NMR active
        |                  one will be used. Default is 1H.

        | Returns:
        |   larmor (float): Larmor frequency in MHz

        """

        el, iso = _el_iso(element)
        return self._B/_larm_units['MHz'](el, iso)

    def set_reference(self, ref, element):
        """
        Set the chemical shift reference (in ppm) for a given element. If not
        provided it will be assumed to be zero.

        | Args:
        |   ref (float): reference shielding value in ppm. Chemical shift will
        |                be calculated as this minus the atom's ms.
        |   element (str): element and isotope whose reference is set.
        |                  Should be in the form <isotope><element>. Isotope
        |                  is optional, if absent the most abundant NMR active
        |                  one will be used.

        """

        el, iso = _el_iso(element)

        if el not in self._references:
            self._references[el] = {}
        self._references[el][iso] = float(ref)

    def set_isotopes(self, isotopes):
        """
        Set the isotopes for each atom in sample.

        | Args:
        |   isotopes (list): list of isotopes for each atom in sample.
        |                    Isotopes can be given as an array of integers or
        |                    of symbols in the form <isotope><element>.
        |                    Their order must match the one of the atoms in
        |                    the original sample ase.Atoms object.
        |                    If an element of the list is None, the most
        |                    common NMR-active isotope is used. If an element
        |                    is the string 'Q', the most common quadrupolar
        |                    active isotope for that nucleus (if known) will
        |                    be used.

        """

        # First: correct length?
        if len(isotopes) != len(self._sample):
            raise ValueError('isotopes array should be as long as the atoms'
                             ' in sample')

        # Clean up the list, make sure it's all right
        iso_clean = []
        for i, iso in enumerate(isotopes):
            # Is it an integer?
            iso_name = ''
            if re.match('[0-9]+', str(iso)) is not None:    # numpy-proof test
                iso_name = str(iso)
                # Does it exist?
                if iso_name not in _nmr_data[self._elems[i]]:
                    raise ValueError('Invalid isotope '
                                     '{0} for element {1}'.format(iso_name,
                                                                  self.
                                                                  _elems[i]))
            elif iso is None:
                iso_name = str(_nmr_data[self._elems[i]]['iso'])
            elif iso == 'Q':
                iso_name = str(_nmr_data[self._elems[i]]['Q_iso'])
            else:
                el, iso_name = _el_iso(iso)
                # Additional test
                if el != self._elems[i]:
                    raise ValueError('Invalid element in isotope array - '
                                     '{0} in place of {1}'.format(el,
                                                                  self
                                                                  ._elems[i]))

            iso_clean.append(iso_name)

        self._isos = np.array(iso_clean)

    def set_element_isotope(self, element, isotope):
        """
        Set the isotope for all occurrences of a given element.

        | Args:
        |   element (str): chemical symbol of the element for which to set the
        |                  isotope.
        |   isotope (int or str): isotope to set for the given element. The
        |                         same conventions as described for the array
        |                         passed to set_isotopes apply.

        """

        new_isos = [int(x) for x in self._isos]
        new_isos = np.where(self._elems == element, isotope, new_isos)

        self.set_isotopes(new_isos)

    def set_single_crystal(self, theta, phi):
        """
        Set the orientation of the sample as a single crystallite.

        | Args:
        |   theta (float): zenithal angle for the crystallite
        |   phi (float): azimuthal angle for the crystallite

        """

        p = np.array([[np.sin(theta)*np.cos(phi),
                       np.sin(theta)*np.sin(phi),
                       np.cos(theta)]])
        w = np.array([1.0])
        t = np.array([])

        self._orients = [p, w, t]

    def set_powder(self, N=8, mode='hemisphere'):
        """
        Set the orientation of the sample as a powder average.

        | Args:
        |   N (int): the number of subdivisions used to generate orientations
        |            within the POWDER algorithm. Higher values make for
        |            better but more expensive averages.
        |   mode (str): which part of the solid angle to cover with the
        |               orientations. Can be 'octant', 'hemisphere' or
        |               'sphere'. The latter should not be necessary for any
        |               NMR interaction. Default is 'hemisphere'.

        """

        self._pwdscheme = TriAvg(mode)
        self._orients = self._pwdscheme.get_orient_points(N)

    def spectrum_1d(self, element, min_freq=-50, max_freq=50, bins=100,
                    freq_broad=None, freq_units='ppm',
                    effects=NMRFlags.CS_ISO, use_central=False,
                    use_reference=False):
        """
        Return a simulated spectrum for the given sample and element.

        | Args:
        |   element (str): element and isotope to get the spectrum of.
        |                  Should be in the form <isotope><element>. Isotope
        |                  is optional, if absent the most abundant NMR active
        |                  one will be used.
        |   min_freq (float): lower bound of the frequency range
        |                    (default is -50)
        |   min_freq (float): upper bound of the frequency range
        |                    (default is 50)
        |   bins (int): number of bins in which to separate the frequency range
        |              (default is 500)
        |   freq_broad (float): Gaussian broadening width to apply to the
        |                       final spectrum (default is None)
        |   freq_units (str): units used for frequency, can be ppm or MHz
        |                     (default is ppm).
        |   effects (NMRFlags): a flag, or bitwise-joined set of flags, from
        |                       this module's NMRFlags tuple, describing which
        |                       effects should be included and accounted for
        |                       in the calculation. For a list of available
        |                       flags check the docstring for NMRCalculator
        |                       (default is NMRFlags.CS_ISO).
        |   use_central (bool): if True, for half-integer spin nuclei, only
        |                       show the central transition. Ignored for
        |                       integer spin nuclei (default is False).
        |   use_reference (bool): if True, return frequencies as referenced to
        |                         the appropriate nucleus, in chemical shift
        |                         form. If no reference has been provided for
        |                         this nucleus, a value of 0 ppm is used and
        |                         the frequencies are simply flipped in sign
        |                         (default is False).

        | Returns:
        |   spec (np.ndarray): array of length 'bins' containing the spectral
        |                      intensities
        |   freq (np.ndarray): array of length 'bins' containing the frequency
        |                      axis

        """

        # First, define the frequency range
        el, iso = _el_iso(element)
        larm = self._B*_nmr_data[el][iso]['gamma']/(2.0*np.pi*1e6)
        I = _nmr_data[el][iso]['I']
        # Units? We want this to be in ppm
        u = {
            'ppm': 1,
            'MHz': 1e6/larm,
        }
        try:
            freq_axis = np.linspace(min_freq, max_freq, bins)*u[freq_units]
        except KeyError:
            raise ValueError('Invalid freq_units passed to spectrum_1d')

        # If it's not a quadrupolar nucleus, no reason to keep those effects
        # around...
        if abs(I) < 1:
            effects &= ~NMRFlags.Q_STATIC
            effects &= ~NMRFlags.Q_MAS

        # Ok, so get the relevant atoms and their properties
        a_inds = np.where((self._elems == el) & (self._isos == iso))[0]

        # Are there even any such atoms?
        if len(a_inds) == 0:
            raise RuntimeError('No atoms of the desired isotopes found in the'
                               ' system')

        # Sanity check
        if (effects & NMRFlags.Q_2_ORIENT_STATIC and
                effects & NMRFlags.Q_2_ORIENT_MAS):
            # Makes no sense...
            raise ValueError('The flags Q_2_ORIENT_STATIC and Q_2_ORIENT_MAS'
                             ' can not be set at the same time')

        if effects & NMRFlags.CS:
            try:
                ms_tens = self._sample.get_array('ms')[a_inds]
                ms_tens = (ms_tens+np.swapaxes(ms_tens, 1, 2))/2.0
            except KeyError:
                raise RuntimeError('Impossible to compute chemical shift - '
                                   'sample has no shielding data')
            ms_evals, ms_evecs = zip(*[np.linalg.eigh(t) for t in ms_tens])

        if effects & (NMRFlags.Q_STATIC | NMRFlags.Q_MAS):
            try:
                efg_tens = self._sample.get_array('efg')[a_inds]
            except KeyError:
                raise RuntimeError('Impossible to compute quadrupolar effects'
                                   ' - sample has no EFG data')
            efg_evals, efg_evecs = zip(*[np.linalg.eigh(t) for t in efg_tens])
            efg_i = (np.arange(len(efg_evals))[:, None],
                     np.argsort(np.abs(efg_evals), axis=1))
            efg_evals = np.array(efg_evals)[efg_i]
            efg_evecs = np.array(efg_evecs)[efg_i[0], :, efg_i[1]]
            Vzz = efg_evals[:, -1]
            eta_q = (efg_evals[:, 0]-efg_evals[:, 1])/Vzz
            Q = _nmr_data[el][iso]['Q']
            chi = Vzz*Q*EFG_TO_CHI

        # Reference (zero if not given)
        try:
            ref = self._references[el][iso]
        except KeyError:
            ref = 0.0

        # Let's start with peak positions - quantities non dependent on
        # orientation

        # Shape: atoms*1Q transitions
        peaks = np.zeros((len(a_inds), int(2*I)))

        # Magnetic quantum number values
        if I % 1 == 0.5 and use_central:
            m = np.array([-0.5, 0.5])[None, :]
        else:
            m = np.arange(-I, I+1).astype(float)[None, :]

        if effects & NMRFlags.CS_ISO:
            peaks += np.average(ms_evals, axis=1)[:, None]

        # Quadrupole second order
        if effects & NMRFlags.Q_2_SHIFT:
            nu_l = larm*1e6
            # NOTE: the last factor of two in this formula was inserted
            # despite not being present in M. J. Duer (5.9) as apparently
            # it's a mistake in the book. Other sources (like the quadrupolar
            # NMR online book by D. Freude and J. Haase, Dec. 2016) report
            # this formulation, with the factor of two, instead.
            q_shifts = np.diff((chi[:, None]/(4*I*(2*I-1)))**2*m/nu_l *
                               (-0.2*(I*(I+1)-3*m**2) *
                                (3+eta_q[:, None]**2))*2)
            q_shifts /= larm

            peaks += q_shifts

        # Any orientational effects at all?
        has_orient = effects & (NMRFlags.CS_ORIENT | NMRFlags.Q_1_ORIENT |
                                NMRFlags.Q_2_ORIENT_STATIC |
                                NMRFlags.Q_2_ORIENT_MAS)
        # Are we using a POWDER average?
        use_pwd = len(self._orients[2]) > 0

        if has_orient:
            # Further expand the peaks!
            peaks = np.repeat(
                peaks[:, :, None], len(self._orients[0]), axis=-1)

        # Now compute the orientational quantities
        if effects & NMRFlags.CS_ORIENT:

            # Compute the traceless ms tensors
            ms_traceless = ms_tens - [np.identity(3)*np.average(ev)
                                      for ev in ms_evals]
            # Now get the shift contributions for each orientation
            dirs = self._orients[0]

            peaks += np.sum(dirs.T[None, :, :]*np.tensordot(ms_traceless,
                                                            dirs,
                                                            axes=((2), (1))),
                            axis=1)[:, None, :]

        if effects & NMRFlags.Q_1_ORIENT:

            # First order quadrupolar anisotropic effects
            # We consider the field aligned along Z
            cosb2 = self._orients[0][:, 2]**2
            sinb2 = 1.0 - cosb2
            cosa2 = (self._orients[0][:, 0]**2) / \
                np.where(sinb2 > 0, sinb2, np.inf)

            dir_fac = 0.5*((3*cosb2[None, :]-1) +
                           eta_q[:, None]*sinb2[None, :]*(2*cosa2[None, :] -
                                                          1.0))
            m_fac = m[:, :-1]+0.5
            nu_q = chi*1.5/(I*(2*I-1.0))

            qfreqs = nu_q[:, None, None]*m_fac[:, :, None]*dir_fac[:, None, :]

            peaks += qfreqs/larm  # Already ppm being Hz/MHz

        if effects & (NMRFlags.Q_2_ORIENT_STATIC | NMRFlags.Q_2_ORIENT_MAS):
            # Which one?
            if effects & NMRFlags.Q_2_ORIENT_STATIC:
                ABC = [_st_A, _st_B, _st_C]
            else:
                ABC = [_mas_A, _mas_B, _mas_C]

            cosa = self._orients[0][:, 0]
            cosb = self._orients[0][:, 1]

            dir_fac = _gfunc(cosa[None, :], cosb[None, :], eta_q[:, None],
                             *ABC)

            m_fac = I*(I+1.0)-17.0/3.0*m[:, :-1]*(m[:, :-1]+1)-13.0/6.0
            nu_q = chi*1.5/(I*(2*I-1.0))

            qfreqs = -((nu_q**2/(6.0*larm*1e6))[:, None, None] *
                       m_fac[:, :, None] *
                       dir_fac[:, None, :])

            peaks += qfreqs/larm

        # Finally, the overall spectrum
        spec = np.zeros(freq_axis.shape)

        for p_nuc in peaks:
            for p_trans in p_nuc:
                if has_orient and use_pwd:
                    spec += self._pwdscheme.average(freq_axis,
                                                    p_trans,
                                                    self._orients[1],
                                                    self._orients[2])

        if freq_broad is None and (not has_orient or not use_pwd):
            print('WARNING: no artificial broadening detected in a calculation'
                  ' without line-broadening contributions. The spectrum could '
                  'appear distorted or empty')

        if freq_broad is not None:
            if has_orient and use_pwd:
                fc = (max_freq+min_freq)/2.0
                bk = np.exp(-((freq_axis-fc)/freq_broad)**2.0)
                bk /= np.sum(bk)
                spec = np.convolve(spec, bk, mode='same')
            else:
                bpeaks = np.exp(-((freq_axis-peaks[:, :, None]) /
                                  freq_broad)**2)  # Broadened peaks
                # Normalise them BY PEAK MAXIMUM
                norm_max = np.amax(bpeaks, axis=-1, keepdims=True)
                norm_max = np.where(np.isclose(norm_max, 0), np.inf, norm_max)
                bpeaks /= norm_max
                spec = np.sum(bpeaks, axis=(0, 1) if not has_orient
                              else (0, 1, 2))

        # Normalize the spectrum to the number of nuclei
        normsum = np.sum(spec)
        if (np.isclose(normsum, 0)):
            print('WARNING: no peaks found in the given frequency range. '
                  'The spectrum will be empty')
        else:
            spec *= len(a_inds)*len(spec)/normsum

        if use_reference:
            freq_axis = ref - freq_axis

        freqs = freq_axis/u[freq_units]
        return spec, freqs

    def dq_buildup(self, sel_i, sel_j=None, t_max=1e-3, t_steps=1000,
                   R_cut=3, kdq=0.155, A=1, tau=np.inf):
        """
        Return a dictionary of double quantum buildup curves for given pairs
        of atoms, built according to the theory given in:

        G. Pileio et al., "Analytical theory of gamma-encoded double-quantum
        recoupling sequences in solid-state nuclear magnetic resonance"
        Journal of Magnetic Resonance 186 (2007) 65-74

        | Args:
        |   sel_i (AtomSelection or [int]): Selection or list of indices of 
        |                                   atoms for which to compute the
        |                                   curves. By default is None
        |                                   (= all of them).
        |   sel_i (AtomSelection or [int]): Selection or list of indices of 
        |                                   atoms for which to compute the
        |                                   curves with sel_i. By default is 
        |                                   None (= same as sel_i).
        |   t_max (float): maximum DQ buildup time, in seconds. Default 
        |                  is 1e-3.
        |   t_steps (int): number of DQ buildup time steps. Default is 1000.
        |   R_cut (float): cutoff radius for which periodic copies to consider
        |                  in each pair, in Angstrom. Default is 3.
        |   kdq (float): same as the k constant in eq. 35 of the reference. A
        |                parameter depending on the specific sequence used.
        |                Default is 0.155. 
        |   A (float): overall scaling factor for the curve. Default is 1.
        |   tau (float): exponential decay factor for the curve. Default
        |                is np.inf.

        | Returns:
        |   curves (dict): a dictionary of all buildup curves indexed by pair,
        |                  plus the time axis in seconds as member 't'.
        """

        tdq = np.linspace(0, t_max, t_steps)

       # Selections
        if sel_i is None:
            sel_i = AtomSelection.all(s)
        elif not isinstance(sel_i, AtomSelection):
            sel_i = AtomSelection(self._sample, sel_i)

        if sel_j is None:
            sel_j = sel_i
        elif not isinstance(sel_j, AtomSelection):
            sel_j = AtomSelection(self._sample, sel_j)

        # Find gammas
        elems = self._sample.get_chemical_symbols()
        gammas = _get_isotope_data(elems, 'gamma', {}, self._isos)

        # Need to sort them and remove any duplicates, also take i < j as
        # convention
        pairs = [tuple(sorted((i, j))) for i in sorted(sel_i.indices)
                 for j in sorted(sel_j.indices)]

        scell_shape = minimum_supcell(R_cut, latt_cart=self._sample.get_cell())
        nfg, ng = supcell_gridgen(self._sample.get_cell(), scell_shape)

        pos = self._sample.get_positions()

        curves = {'t': tdq}

        for ij in pairs:

            r = pos[ij[1]]-pos[ij[0]]

            all_r = r[None, :]+ng

            all_R = np.linalg.norm(all_r, axis=1)

            # Apply cutoff
            all_R = all_R[np.where((all_R <= R_cut)*(all_R > 0))]
            n = all_R.shape[0]

            bij = _dip_constant(all_R*1e-10, gammas[ij[0]], gammas[ij[1]])

            th = 1.5*kdq*abs(bij[:, None])*tdq[None, :]*2*np.pi
            x = (2*th/np.pi)**0.5

            Fs, Fc = fresnel(x*2**0.5)

            x[:, 0] = np.inf

            bdup = 0.5-(1.0/(x*8**0.5)) * \
                (Fc*np.cos(2*th) + Fs*np.sin(2*th))
            bdup[:, 0] = 0

            curves[ij] = A*np.sum(bdup, axis=0)*np.exp(-tdq/tau)

        return curves

    @property
    def B(self):
        """Static magnetic field, in Tesla"""
        return self._B



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
shrewd.py

Contains a class to define the SHREWD (Spherical Harmonics Reduction or Elimination by 
a Weighted Distribution) powder averaging scheme.
Implementation taken from:

Mattias Edén, Malcolm H. Levitt,
"Computation of Orientational Averages in Solid-State NMR by Gaussian Spherical Quadrature",
Journal of Magnetic Resonance,
Volume 132, Issue 2,
1998,
Pages 220-239,
ISSN 1090-7807,
https://doi.org/10.1006/jmre.1998.1427.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import warnings
from scipy.special import legendre, sph_harm
from scipy.linalg import lstsq
from scipy.optimize import minimize
from soprano.calculate.powder.zcw import ZCW


class SHREWD(ZCW):

    def _calc_engine(self, N):
        phi, ct, weights = super()._calc_engine(N)

        # Compute weights
        lmax = len(phi)
        lM = np.array([legendre(l)(ct) for l in range(lmax)]).T
        lb = np.array([1] + [0]*(lmax-1))

        alpha = 0.1 # This term helps with keeping the weights in check

        def f(w):
            return np.sum((np.dot(lM, w)-lb)**2 + alpha*w**2)

        def df(w):
            return 2*(np.dot((np.dot(lM, w)-lb), lM)) + 2*alpha*w

        sol = minimize(f, weights, jac=df)
        if sol.status != 0:
            warnings.warn('Optimization of weights for SHREWD did not converge')
        weights = sol.x
        # Normalise them, just for safety
        weights /= np.sum(weights)

        return phi, ct, weights



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Definition of a fake QueueInterface class, useful for debugging Submitters.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import time
import numpy as np

import subprocess as sp
from threading import Thread
from soprano.utils import safe_communicate
from soprano.hpc.submitter.queues import QueueInterface


class DebugQueueInterface(QueueInterface):

    """DebugQueueInterface object

    A class meant to emulate a QueueInterface while doing absolutely nothing
    of what it does. Jobs are simply stored locally, there's a fixed waiting
    time, and are then executed. Ideally they should be simple, quick stuff
    (like an echo command). No guarantees for actually long jobs.

    In the submitted script a syntax for additional variables is allowed,
    similar to real queue systems. These follow the convention of many engines
    of having to start with #$. For example

    #$ WAIT 10

    means the job will be put in a "wait" state for 10 seconds. The currently
    available variables are:

    WAIT - specify how long the job has to stay in a "wait" state. If two
    values are provided, these are considered bounds for a random number
    RUN - same as above, but for the running state. This has no bearing on the
    *actual* running time (it's suggested that it's something very quick)

    """

    def __init__(self, dt=0.1):
        """Initialize the DebugQueueInterface.

        | Args:
        |   dt (float): frequency with which the queue status is updated

        """

        self._job_list = {}
        self._dt = dt

        self._runthr = None

    def _main_loop(self):

        while(len(self._job_list) > 0):
            t = time.time()
            completed_jobs = []
            for j_id in list(self._job_list.keys()):
                job = self._job_list[j_id]
                if job['status'] == 'w' and (t-job['t0']) > job['WAIT']:
                    # Switch to run
                    job['status'] = 'r'
                    # And actually run it!
                    proc = sp.Popen(['bash'], stdin=sp.PIPE,
                                    stdout=sp.PIPE,
                                    stderr=sp.PIPE,
                                    cwd=job['cwd'])
                    stdout, stderr = safe_communicate(proc, job['script'])
                elif job['status'] == 'r' \
                        and (t-job['t0']-job['WAIT']) > job['RUN']:
                    # Just eliminate it
                    completed_jobs.append(j_id)

            for cj in completed_jobs:
                del(self._job_list[cj])

            time.sleep(self._dt)

    def submit(self, script, cwd=None):
        """Submit a job to the queue.

        | Args:
        |   script (str): content of the submission script
        |   cwd (Optional[str]): path to the desired working directory
        |
        | Returns:
        |   job_id (str): the job ID assigned by the queue system and parsed
        |                 with sub_outre
        """

        # Parse the script for parameters
        job = {
            'WAIT': 0.0,
            'RUN': 0.0
        }

        for l in script.split('\n'):
            if l[:2] == '#$':
                # Parse!
                keyw, vals = l[2:].split(None, 1)
                if keyw in job:
                    if type(job[keyw]) is float:
                        # Ok, what do we have?
                        vals = vals.split()
                        if len(vals) == 1:
                            job[keyw] = float(vals[0])
                        elif len(vals) == 2:
                            job[keyw] = np.random.uniform(*map(float, vals))

        # Python 2-to-3 compatibility
        try:
            script = bytes(script, 'utf-8')
        except TypeError:
            pass

        # Ok, now add the actual script
        job['script'] = script
        job['cwd'] = cwd
        job['t0'] = time.time()
        job['status'] = 'w'
        # Generate a suitable id
        new_id = '1'
        while new_id in self._job_list:
            new_id = str(np.random.randint(10**(len(new_id)),
                                           10**(1+len(new_id))))

        self._job_list[new_id] = job
        if self._runthr is None or not(self._runthr.isAlive()):
            self._runthr = Thread(target=self._main_loop)
            self._runthr.daemon = True
            self._runthr.start()

        return new_id

    def list(self):
        """List all jobs found in the queue

        | Returns:
        |   jobs (dict): a dict of jobs classified by ID containing all info
        |                that can be matched through list_outre
        |
        """

        jobs = {}
        for j_id in self._job_list:
            jobs[j_id] = {'job_id': j_id,
                          'job_status': self._job_list[j_id]['status']}

        return jobs

    def kill(self, job_id):
        """Kill the job with the given ID

        | Args:
        |   job_id (str): ID of the job to kill
        |
        """

        try:
            del(self._job_list[job_id])
        except KeyError:
            pass
